# mypy: disable-error-code=misc
import asyncio
import sys
import time
from http import HTTPStatus
from typing import Any, Optional

import structlog
from sanic import Blueprint, HTTPResponse, response
from sanic.request import Request
from sanic_openapi import openapi

import rasa
from rasa.builder.auth import HEADER_USER_ID, is_auth_required_now, protected
from rasa.builder.config import (
    COPILOT_ASSISTANT_TRACKER_MAX_TURNS,
    COPILOT_HANDLER_ROLLING_BUFFER_SIZE,
    GUARDRAILS_ENABLE_BLOCKING,
    HELLO_RASA_PROJECT_ID,
    LAKERA_ASSISTANT_HISTORY_GUARDRAIL_PROJECT_ID,
    LAKERA_COPILOT_HISTORY_GUARDRAIL_PROJECT_ID,
)
from rasa.builder.copilot.constants import ROLE_USER, SIGNATURE_VERSION_V1
from rasa.builder.copilot.copilot_response_handler import CopilotResponseHandler
from rasa.builder.copilot.exceptions import (
    CopilotStreamError,
    InvalidCopilotChatHistorySignature,
    MissingCopilotChatHistorySignature,
)
from rasa.builder.copilot.models import (
    CopilotContext,
    CopilotRequest,
    GeneratedContent,
    ReferenceEntry,
    ReferenceSection,
    ResponseCategory,
    ResponseCompleteness,
)
from rasa.builder.copilot.signing import (
    create_signature_envelope_for_handler,
    create_signature_envelope_for_text,
    verify_signature,
)
from rasa.builder.copilot.telemetry import CopilotTelemetry
from rasa.builder.download import create_bot_project_archive
from rasa.builder.guardrails.constants import (
    BLOCK_SCOPE_PROJECT,
    BLOCK_SCOPE_USER,
    BlockScope,
)
from rasa.builder.guardrails.store import guardrails_store
from rasa.builder.job_manager import job_manager
from rasa.builder.jobs import (
    run_prompt_to_bot_job,
    run_replace_all_files_job,
    run_template_to_bot_job,
)
from rasa.builder.llm_service import llm_service
from rasa.builder.logging_utils import (
    capture_exception_with_context,
    get_recent_logs,
)
from rasa.builder.models import (
    AgentStatus,
    ApiErrorResponse,
    AssistantInfo,
    BotData,
    JobCreateResponse,
    JobStatus,
    JobStatusEvent,
    PromptRequest,
    ServerSentEvent,
    TemplateRequest,
)
from rasa.builder.project_generator import ProjectGenerator
from rasa.builder.shared.tracker_context import TrackerContext
from rasa.core.agent import Agent
from rasa.core.channels.studio_chat import StudioChatInput
from rasa.core.exceptions import AgentNotReady
from rasa.shared.core.flows.flows_list import FlowsList
from rasa.shared.core.flows.yaml_flows_io import get_flows_as_json
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.importers.utils import DOMAIN_KEYS
from rasa.utils.json_utils import extract_values
from rasa.utils.openapi import model_to_schema

structlogger = structlog.get_logger()

# Create the blueprint
bp = Blueprint("bot_builder", url_prefix="/api")


def setup_project_generator(project_folder: str) -> ProjectGenerator:
    """Initialize and return a ProjectGenerator instance."""
    # Ensure the project folder is in sys.path
    if project_folder not in sys.path:
        sys.path.insert(0, project_folder)

    structlogger.info(
        "bot_builder_service.service_initialized", project_folder=project_folder
    )

    return ProjectGenerator(project_folder)


def get_project_generator(request: Request) -> ProjectGenerator:
    """Get the project generator from app context."""
    return request.app.ctx.project_generator


def get_input_channel(request: Request) -> StudioChatInput:
    """Get the input channel from app context."""
    return request.app.ctx.input_channel


async def extract_bot_data_from_agent(agent: Agent) -> BotData:
    """Extract BotData from an Agent.

    Args:
        agent: The agent to extract data from

    Returns:
        BotData containing flows, domain, config, endpoints, and nlu data
    """
    domain = agent.domain.as_dict() if agent.domain else {}
    flows = (
        await agent.processor.get_flows()
        if agent.processor
        else FlowsList(underlying_flows=[])
    )
    return BotData(
        flows=get_flows_as_json(flows),
        domain=extract_values(domain, DOMAIN_KEYS),
    )


async def get_agent_status(request: Request) -> AgentStatus:
    """Get the status of the agent."""
    if request.app.ctx.agent is None:
        return AgentStatus.not_loaded
    agent: Agent = request.app.ctx.agent
    if agent.is_ready():
        return AgentStatus.ready
    return AgentStatus.not_ready


# Health check endpoint
@bp.route("/", methods=["GET"])
@openapi.summary("Health check endpoint")
@openapi.description(
    "Returns the health status of the Bot Builder service including version "
    "information and authentication requirements"
)
@openapi.tag("health")
@openapi.response(
    200,
    {
        "application/json": {
            "status": str,
            "service": str,
            "rasa_version": str,
            "auth_required": bool,
            "agent_status": str,
        }
    },
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
async def health(request: Request) -> HTTPResponse:
    """Health check endpoint."""
    project_generator = get_project_generator(request)
    return response.json(
        {
            "status": "ok",
            "service": "bot-builder",
            "rasa_version": rasa.__version__,
            "agent_status": await get_agent_status(request),
            "auth_required": is_auth_required_now(
                project_info=project_generator.project_info
            ),
        }
    )


@bp.route("/job-events/<job_id>", methods=["GET"])
@openapi.summary("Stream job progress events")
@openapi.description(
    "Stream server-sent events (SSE) tracking real-time job progress.\n\n"
    "**Connect with:** `Accept: text/event-stream`.\n\n"
    "**SSE Event Example:**\n"
    "```text\n"
    "event: received\n"
    'data: {"status": "received"}\n'
    "```\n\n"
)
@openapi.tag("job-events")
@openapi.parameter(
    "job_id", str, location="path", description="The id of the job to stream events for"
)
@openapi.response(
    200,
    {"text/event-stream": str},
    description="Server-sent events stream. See documentation for event types.",
    example=(
        "event: received\n"
        'data: {"status": "received"}\n'
        "\n"
        "event: generating\n"
        'data: {"status": "generating"}\n'
    ),
)
@openapi.response(
    404,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Unknown job_id: No such job exists.",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def job_events(request: Request, job_id: str) -> HTTPResponse:
    try:
        job = job_manager.get_job(job_id)
        if job is None:
            return response.json(
                ApiErrorResponse(
                    error="Job not found", details={"job_id": job_id}
                ).model_dump(),
                status=404,
            )

        stream = await request.respond(content_type="text/event-stream")

        try:
            async for evt in job.event_stream():
                await stream.send(evt.format())
        except Exception as exc:
            # Handle exceptions within the SSE stream context
            capture_exception_with_context(
                exc,
                "bot_builder_service.job_events.streaming_error",
                extra={"job_id": job_id},
                tags={"endpoint": "/api/job-events/<job_id>"},
            )
            # Send error event in SSE format instead of JSON response
            error_event = JobStatusEvent.from_status(
                status=JobStatus.error.value,
                message=f"Failed to stream job events: {exc}",
            ).format()
            await stream.send(error_event)
        finally:
            await stream.eof()

        return stream
    except Exception as exc:
        # This exception handler only applies before stream.respond() is called
        capture_exception_with_context(
            exc,
            "bot_builder_service.job_events.unexpected_error",
            extra={"job_id": job_id},
            tags={"endpoint": "/api/job-events/<job_id>"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to stream job events", details={"error": str(exc)}
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/prompt-to-bot", methods=["POST"])
@openapi.summary("Generate bot from natural language prompt")
@openapi.description(
    "Creates a complete conversational AI bot from a natural language prompt. "
    "Returns immediately with a job ID. Connect to `/job-events/<job_id>` to "
    "receive server-sent events (SSE) for real-time progress tracking "
    "throughout the bot creation process.\n\n"
    "**SSE Event Flow** (via `/job-events/<job_id>`):\n"
    "1. `received` - Request received by server\n"
    "2. `generating` - Generating bot project files\n"
    "3. `generation_success` - Bot generation completed successfully\n"
    "4. `training` - Training the bot model\n"
    "5. `train_success` - Model training completed\n"
    "6. `done` - Bot creation completed\n\n"
    "**Error Events:**\n"
    "- `generation_error` - Failed to generate bot from prompt\n"
    "- `train_error` - Bot generated but training failed\n"
    "- `validation_error` - Generated bot configuration is invalid\n"
    "- `error` - Unexpected error occurred\n\n"
    "**Usage:**\n"
    "1. Send POST request with Content-Type: application/json\n"
    "2. The response will be a JSON object `{job_id: ...}`\n"
    "3. Connect to `/job-events/<job_id>` for a server-sent event stream of progress."
)
@openapi.tag("bot-generation")
@openapi.body(
    {"application/json": model_to_schema(PromptRequest)},
    description="Prompt request with natural language description.",
    required=True,
    example={
        "prompt": (
            "Create a customer support bot that can help users with order inquiries, "
            "product questions, and returns processing. The bot should be friendly "
            "and able to escalate to human agents when needed."
        )
    },
)
@openapi.response(
    200,
    {"application/json": model_to_schema(JobCreateResponse)},
    description="Job created. Poll or subscribe to /job-events/<job_id> for progress.",
)
@openapi.response(
    400,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Validation error in request payload",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def handle_prompt_to_bot(request: Request) -> HTTPResponse:
    """Handle prompt-to-bot generation requests."""
    try:
        payload = PromptRequest(**request.json)
    except Exception as exc:
        return response.json(
            ApiErrorResponse(
                error="Invalid request", details={"error": str(exc)}
            ).model_dump(),
            status=400,
        )

    try:
        # Allocate job and schedule background task
        job = job_manager.create_job()
        request.app.add_task(run_prompt_to_bot_job(request.app, job, payload.prompt))
        return response.json(JobCreateResponse(job_id=job.id).model_dump(), status=200)
    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.prompt_to_bot.unexpected_error",
            tags={"endpoint": "/api/prompt-to-bot"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to create prompt-to-bot job",
                details={"error": str(exc)},
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/template-to-bot", methods=["POST"])
@openapi.summary("Generate bot from predefined template")
@openapi.description(
    "Creates a complete conversational AI bot from a predefined template. "
    "Returns immediately with a job ID. Connect to `/job-events/<job_id>` to "
    "receive server-sent events (SSE) for real-time progress tracking "
    "throughout the bot creation process.\n\n"
    "**SSE Event Flow** (via `/job-events/<job_id>`):\n"
    "1. `received` - Request received by server\n"
    "2. `generating` - Initializing bot from template\n"
    "3. `generation_success` - Template initialization completed successfully\n"
    "4. `training` - Training the bot model\n"
    "5. `train_success` - Model training completed\n"
    "6. `done` - Bot creation completed\n\n"
    "**Error Events:**\n"
    "- `generation_error` - Failed to initialize bot from template\n"
    "- `train_error` - Template loaded but training failed\n"
    "- `validation_error` - Template configuration is invalid\n"
    "- `error` - Unexpected error occurred\n\n"
    "**Usage:**\n"
    "1. Send POST request with Content-Type: application/json\n"
    "2. The response will be a JSON object `{job_id: ...}`\n"
    "3. Connect to `/job-events/<job_id>` for a server-sent event stream of progress."
)
@openapi.tag("bot-generation")
@openapi.body(
    {"application/json": model_to_schema(TemplateRequest)},
    description="Template request with template name.",
    required=True,
    example={"template_name": "telco"},
)
@openapi.response(
    200,
    {"application/json": model_to_schema(JobCreateResponse)},
    description="Job created. Poll or subscribe to /job-events/<job_id> for progress.",
)
@openapi.response(
    400,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Validation error in request payload or invalid template name",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def handle_template_to_bot(request: Request) -> HTTPResponse:
    """Create a new template-to-bot job and return job_id immediately."""
    try:
        template_data = TemplateRequest(**request.json)
    except Exception as exc:
        return response.json(
            ApiErrorResponse(
                error="Invalid request", details={"error": str(exc)}
            ).model_dump(),
            status=400,
        )

    try:
        # allocate job and schedule background task
        job = job_manager.create_job()
        request.app.add_task(
            run_template_to_bot_job(request.app, job, template_data.template_name)
        )
        return response.json(JobCreateResponse(job_id=job.id).model_dump(), status=200)
    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.template_to_bot.unexpected_error",
            tags={"endpoint": "/api/template-to-bot"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to create template-to-bot job",
                details={"error": str(exc)},
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/files", methods=["GET"])
@openapi.summary("Get bot files")
@openapi.description(
    "Retrieves the current bot configuration files including domain.yml, "
    "config.yml, flows.yml, NLU data, and other project files as a "
    "dictionary mapping file names to their string contents"
)
@openapi.tag("bot-files")
@openapi.response(
    200,
    {"application/json": {str: Optional[str]}},
    description="Bot files retrieved successfully",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def get_bot_files(request: Request) -> HTTPResponse:
    """Get current bot files."""
    try:
        project_generator = get_project_generator(request)
        bot_files = project_generator.get_bot_files()
        return response.json(bot_files)
    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.get_bot_files.unexpected_error",
            tags={"endpoint": "/api/files", "method": "GET"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to retrieve bot files", details={"error": str(exc)}
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/files", methods=["POST"])
@openapi.summary("Replace all bot files")
@openapi.description(
    "Replaces all bot configuration files with the provided files, deletes any "
    "files not included in the request (excluding .rasa/ and models/ directories), "
    "and retrains the model. Returns immediately with a job ID. Connect to "
    "`/job-events/<job_id>` for real-time SSE progress tracking."
    "\n\n"
    "**File Management:**\n"
    "- All files in the request are written to the project folder\n"
    "- Files not included in the request are deleted from the project\n"
    "- Files/folders starting with `.rasa/` or `models/` are excluded from deletion\n\n"
    "**SSE Event Flow:** (available via /job-events/<job_id>)\n"
    "1. `received` - Request received by server\n"
    "2. `validating` - Validating bot configuration files\n"
    "3. `validation_success` - File validation completed successfully\n"
    "4. `training` - Training the bot model with updated files\n"
    "5. `train_success` - Model training completed\n"
    "6. `done` - Bot files replacement completed\n\n"
    "**Error Events (can occur at any time):**\n"
    "- `validation_error` - Bot configuration files are invalid\n"
    "- `train_error` - Files updated but training failed\n"
    "- `copilot_analysis_start` - Copilot analysis started (includes `copilot_job_id` "
    "in payload)\n"
    "- `error` - Unexpected error occurred\n\n"
    "**Copilot Analysis Events:**\n"
    "When training or validation fails, a separate copilot analysis job is "
    "automatically started. The `copilot_analysis_start` event includes a "
    "`copilot_job_id` in the payload.\nConnect to `/job-events/<copilot_job_id>` to "
    "receive the following events:\n"
    "- `copilot_analyzing` - Copilot is analyzing errors and providing suggestions. "
    "Uses the same SSE event payload format as the `/copilot` endpoint with `content`, "
    "`response_category`, and `completeness` fields.\n"
    "- `copilot_analysis_success` - Copilot analysis completed with references.\n"
    "- `copilot_analysis_error` - Copilot analysis failed\n\n"
    "**Usage:**\n"
    "1. Send POST request with Content-Type: application/json\n"
    "2. The response will be a JSON object `{job_id: ...}`\n"
    "3. Connect to `/job-events/<job_id>` for a server-sent event stream of progress."
)
@openapi.tag("bot-files")
@openapi.body(
    {"application/json": {str: Optional[str]}},
    description=(
        "A dictionary mapping file names to their complete content. "
        "All files in the project will be replaced with these files. "
        "Files not included in the request will be deleted from the project "
        "(except for .rasa/ and models/ directories). "
        "The file name should be the relative path from the project root."
    ),
    required=True,
    example={
        "domain.yml": (
            "version: '3.1'\n"
            "intents:\n  - greet\n  - goodbye\n"
            "responses:\n  utter_greet:\n  - text: 'Hello!'\n"
            "  utter_goodbye:\n  - text: 'Goodbye!'"
        ),
        "config.yml": (
            "version: '3.1'\n"
            "pipeline:\n  - name: WhitespaceTokenizer\n"
            "  - name: RegexFeaturizer\n  - name: LexicalSyntacticFeaturizer\n"
            "  - name: CountVectorsFeaturizer\n"
            "policies:\n  - name: MemoizationPolicy\n  - name: RulePolicy"
        ),
        "data/nlu.yml": (
            "version: '3.1'\n"
            "nlu:\n- intent: greet\n  examples: |\n    - hello\n    - hi"
        ),
    },
)
@openapi.response(
    200,
    {"application/json": model_to_schema(JobCreateResponse)},
    description=(
        "Job created. Poll or subscribe to /job-events/<job_id> "
        "for progress and SSE updates on file replacement and training."
    ),
)
@openapi.response(
    400,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Validation error in bot files",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.response(
    401,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description=(
        "Authentication failed - Authorization header missing or invalid. "
        "Auth may be conditionally required depending on server "
        "configuration."
    ),
)
@openapi.parameter(
    "Authorization",
    description=(
        "Bearer token for authentication. Required after auth start window "
        "or if configured."
    ),
    _in="header",
    required=False,
    schema=str,
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
@protected()
async def replace_all_bot_files(request: Request) -> HTTPResponse:
    """Replace all bot files with server-sent events for progress tracking."""
    try:
        bot_files = request.json
    except Exception as exc:
        return response.json(
            ApiErrorResponse(
                error="Invalid request", details={"error": str(exc)}
            ).model_dump(),
            status=400,
        )

    try:
        job = job_manager.create_job()
        request.app.add_task(run_replace_all_files_job(request.app, job, bot_files))
        return response.json(JobCreateResponse(job_id=job.id).model_dump(), status=200)
    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.replace_all_bot_files.unexpected_error",
            tags={"endpoint": "/api/files", "method": "POST"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to replace bot files", details={"error": str(exc)}
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/data", methods=["GET"])
@openapi.summary("Get bot data")
@openapi.description(
    "Retrieves the current bot data in CALM import format with flows, domain, "
    "config, endpoints, and NLU data"
)
@openapi.tag("bot-info")
@openapi.response(
    200,
    {"application/json": model_to_schema(BotData)},
    description="Bot data retrieved successfully",
)
@openapi.response(
    409,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Agent not ready",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def get_bot_data(request: Request) -> HTTPResponse:
    """Get current bot data in CALM import format."""
    try:
        agent: Optional[Agent] = request.app.ctx.agent
        if not agent:
            raise AgentNotReady(
                "Can't retrieve the data without an agent being loaded."
            )

        bot_data = await extract_bot_data_from_agent(agent)

        return response.json(bot_data.model_dump())
    except AgentNotReady as e:
        return response.json(
            ApiErrorResponse(
                error="Agent not ready",
                details={"error": str(e)},
            ).model_dump(),
            status=HTTPStatus.CONFLICT,
        )
    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.get_bot_data.unexpected_error",
            tags={"endpoint": "/api/data"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to retrieve bot data",
                details={"error": str(exc)},
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/assistant", methods=["GET"])
@openapi.summary("Get assistant info")
@openapi.description(
    "Returns basic information about the loaded assistant, including the assistant id "
    "as configured in the model's metadata (from config.yml)."
)
@openapi.tag("bot-info")
@openapi.response(
    200,
    {"application/json": model_to_schema(AssistantInfo)},
    description="Assistant info retrieved successfully",
)
@openapi.response(
    409,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Agent not ready",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def get_bot_info(request: Request) -> HTTPResponse:
    """Return assistant info including assistant id from model metadata."""
    try:
        agent: Optional[Agent] = request.app.ctx.agent
        if not agent:
            raise AgentNotReady("Can't retrieve bot info without a loaded agent.")

        assistant_id: Optional[str] = (
            agent.processor.model_metadata.assistant_id
            if agent.processor
            and agent.processor.model_metadata
            and hasattr(agent.processor.model_metadata, "assistant_id")
            else None
        )

        return response.json(AssistantInfo(assistant_id=assistant_id).model_dump())
    except AgentNotReady as e:
        return response.json(
            ApiErrorResponse(
                error="Agent not ready",
                details={"error": str(e)},
            ).model_dump(),
            status=HTTPStatus.CONFLICT,
        )
    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.get_bot_info.unexpected_error",
            tags={"endpoint": "/api/assistant"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to retrieve bot info",
                details={"error": str(exc)},
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/download", methods=["GET"])
@openapi.summary("Download bot project as tar.gz")
@openapi.description(
    "Downloads the current bot project files as a compressed tar.gz archive. "
    "Includes all configuration files and a .env file with RASA_PRO_LICENSE. "
    "Requires valid JWT token in Authorization header."
)
@openapi.tag("bot-files")
@openapi.parameter(
    "Authorization",
    description=("Bearer token for authentication. Always required for this endpoint."),
    _in="header",
    required=True,
    schema=str,
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
@openapi.parameter(
    "project_name",
    description="Name of the project for the archive filename and pyproject.toml",
    _in="query",
    required=False,
    schema=str,
)
@openapi.response(
    200,
    {"application/gzip": bytes},
    description="Bot project downloaded successfully as tar.gz",
)
@openapi.response(
    401,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description=(
        "Authentication failed - Authorization header missing or invalid. "
        "Authentication is always required for this endpoint."
    ),
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@protected(always_required=True)
async def download_bot_project(request: Request) -> HTTPResponse:
    """Download bot project as tar.gz archive."""
    try:
        # Token verification is enforced by the
        # protected(always_required=True) decorator.

        # Get bot files
        project_generator = get_project_generator(request)
        bot_files = project_generator.get_bot_files()

        # Get project name from query parameters, default to "bot-project"
        project_name = request.args.get("project_name", "bot-project")

        # Create tar.gz archive
        tar_data = create_bot_project_archive(bot_files, project_name)

        structlogger.info(
            "bot_builder_service.download_bot_project.success",
            user_sub=(getattr(request.ctx, "auth_payload", None) or {}).get("sub"),
            files_count=len(bot_files),
            archive_size=len(tar_data),
            payload=getattr(request.ctx, "auth_payload", None),
            project_name=project_name,
        )

        return response.raw(
            tar_data,
            content_type="application/gzip",
            headers={
                "Content-Disposition": f"attachment; filename={project_name}.tar.gz"
            },
        )

    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.download_bot_project.unexpected_error",
            tags={"endpoint": "/api/download"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to create bot project archive",
                details={"error": str(exc)},
            ).model_dump(),
            status=500,
        )


@bp.route("/copilot", methods=["POST"])
@openapi.summary("AI copilot for bot building")
@openapi.description(
    "Provides LLM-powered copilot assistance with streaming markdown responses. "
    "Returns server-sent events (SSE) for real-time streaming of copilot responses.\n\n"
    "The event's `event` field is the type of event. For this endpoint it will always "
    "be: `copilot_response`.\n\n"
)
@openapi.tag("copilot")
@openapi.body(
    {"application/json": model_to_schema(CopilotRequest)},
    description=(
        "Copilot request containing: "
        "1. conversation history between user and copilot, "
        "2. session ID for tracking conversation context with the bot being built."
    ),
    required=True,
)
@openapi.response(
    200,
    {"text/event-stream": model_to_schema(ServerSentEvent)},
    description=(
        "Server-sent events stream with copilot responses and references. "
        "The event's `event` field is the type "
        "of event. For this endpoint it will always be: `copilot_response`. "
        "The event's data field is a JSON dump. The following describes the event data "
        "field:\n"
        "- `response_category` can be one of the following:\n"
        "  - `copilot` - Stream token generated by the copilot.\n"
        "  - `out_of_scope_detection` - Response coming from the the out of scope detection.\n"  # noqa: E501
        "  - `roleplay_detection` - Response coming from the the roleplay detection.\n"
        "  - `reference` - Reference section.\n"
        "  - `training_error_log_analysis` - Used to flag the responses that are training error log analysis.\n"  # noqa: E501
        "  - `e2e_testing_error_log_analysis` - Used to flag the responses that are e2e testing error log analysis.\n"  # noqa: E501
        "  - `guardrail_policy_violation` - Used to flag the responses that are flagged as a violation of the guardrail policy.\n\n"  # noqa: E501
        "- `completeness`: Whether this is a streaming token or complete response.\n\n"
        "- `content`: The actual response content (for streaming tokens).\n\n"
        "- `references`: (Only for `reference` response category) List of reference entries. Each reference entry is a dictionary with the following keys:\n\n"  # noqa: E501
        "   - `index`: Reference index as an integer number.\n"
        "   - `title`: Reference title as a string.\n"
        "   - `url`: Reference URL as a string.\n\n"
    ),
    examples={
        "Streaming token from copilot": {
            "summary": "Streaming token from copilot",
            "value": GeneratedContent(
                content="<token generated by the copilot>",
                response_category=ResponseCategory.COPILOT,
                response_completeness=ResponseCompleteness.TOKEN,
            )
            .to_sse_event()
            .model_dump(),
        },
        "Complete response for the following response categories: out_of_scope_detection / roleplay_detection / training_error_log_analysis / e2e_testing_error_log_analysis / guardrail_policy_violation": {  # noqa: E501
            "summary": "Complete response from copilot",
            "value": GeneratedContent(
                content="<response from the out of scope detection>",
                response_category=ResponseCategory.OUT_OF_SCOPE_DETECTION,
                response_completeness=ResponseCompleteness.COMPLETE,
            )
            .to_sse_event()
            .model_dump(),
        },
        "Response with the references": {
            "summary": "Reference section with all references",
            "value": ReferenceSection(
                references=[
                    ReferenceEntry(
                        index=1,
                        title="Title of the reference",
                        url="https://rasa.com/docs/...",
                    ),
                    ReferenceEntry(
                        index=2,
                        title="Title of another reference",
                        url="https://rasa.com/docs/...",
                    ),
                ],
                response_category=ResponseCategory.REFERENCE,
                response_completeness=ResponseCompleteness.COMPLETE,
            )
            .to_sse_event()
            .model_dump(),
        },
    },
)
@openapi.response(
    400,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Validation error in request",
)
@openapi.response(
    502,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="LLM generation failed.",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error.",
)
@openapi.response(
    401,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description=(
        "Authentication failed - Authorization header missing or invalid. "
        "Auth may be conditionally required depending on server "
        "configuration."
    ),
)
@openapi.parameter(
    "Authorization",
    description=(
        "Bearer token for authentication. Required after auth start window "
        "or if configured."
    ),
    _in="header",
    required=False,
    schema=str,
)
@openapi.parameter(
    HEADER_USER_ID,
    description=("Required user id used for telemetry and guardrails (X-User-Id)."),
    _in="header",
    required=True,
    schema=str,
)
@protected()
async def copilot(request: Request) -> None:
    """Handle copilot requests with streaming markdown responses."""
    sse = await request.respond(content_type="text/event-stream")
    project_generator = get_project_generator(request)

    try:
        # 1. Validate and unpack input
        req = CopilotRequest(**request.json)

        # Require user identifier via header and fail fast if missing
        user_id = (request.headers.get(HEADER_USER_ID) or "").strip()
        if not user_id:
            structlogger.error(
                "bot_builder_service.copilot.missing_or_empty_user_id_header",
                event_info=f"Missing or empty required header: {HEADER_USER_ID}",
            )
            await sse.send(
                ServerSentEvent(
                    event="error",
                    data={
                        "error": f"Missing or empty required header: {HEADER_USER_ID}"
                    },
                ).format()
            )
            return

        telemetry = CopilotTelemetry(project_id=HELLO_RASA_PROJECT_ID, user_id=user_id)
        structlogger.debug("builder.copilot.telemetry.request.init")

        if req.last_message and req.last_message.role == ROLE_USER:
            structlogger.debug("builder.copilot.telemetry.request.user_turn")
            # Offload telemetry logging to a background task
            request.app.add_task(
                asyncio.to_thread(
                    telemetry.log_user_turn,
                    req.last_message.get_flattened_text_content(),
                )
            )

        # 2. Check if we need to block the request due to too many guardrails violations
        if (scope := await _get_copilot_block_scope(user_id)) is not None:
            message = CopilotResponseHandler.respond_to_guardrail_blocked(scope)
            await sse.send(message.to_sse_event().format())
            return

        # 3 Verify the request signature
        try:
            await verify_signature(req)
        except InvalidCopilotChatHistorySignature:
            version = getattr(req, "signature_version", None) or SIGNATURE_VERSION_V1
            await sse.send(
                ServerSentEvent(
                    event="copilot_response",
                    data={"error": "invalid_history_signature", "version": version},
                ).format()
            )
            return
        except MissingCopilotChatHistorySignature:
            version = getattr(req, "signature_version", None) or SIGNATURE_VERSION_V1
            await sse.send(
                ServerSentEvent(
                    event="copilot_response",
                    data={"error": "missing_history_signature", "version": version},
                ).format()
            )
            return

        # 4. Get the necessary context for the copilot
        tracker = await current_tracker_from_input_channel(request.app, req.session_id)
        tracker_context = TrackerContext.from_tracker(
            tracker, max_turns=COPILOT_ASSISTANT_TRACKER_MAX_TURNS
        )
        if (
            tracker_context is not None
            and llm_service.guardrails_policy_checker is not None
        ):
            tracker_context = await llm_service.guardrails_policy_checker.check_assistant_chat_for_policy_violations(  # noqa: E501
                tracker_context=tracker_context,
                hello_rasa_user_id=user_id,
                hello_rasa_project_id=HELLO_RASA_PROJECT_ID,
                lakera_project_id=LAKERA_ASSISTANT_HISTORY_GUARDRAIL_PROJECT_ID,
            )

        # Copilot doesn't need to know about the docs and any file that is not a core
        # assistant file
        relevant_assistant_files = project_generator.get_bot_files(
            exclude_docs_directory=True,
            allowed_file_extensions=["yaml", "yml", "py", "jinja", "jinja2"],
        )
        context = CopilotContext(
            tracker_context=tracker_context,
            assistant_logs=get_recent_logs(),
            assistant_files=relevant_assistant_files,
            copilot_chat_history=req.copilot_chat_history,
        )

        # 5. Run guardrail policy checks. If any policy violations are detected,
        #    send a response and end the stream.
        guardrail_response: Optional[GeneratedContent] = None
        if llm_service.guardrails_policy_checker is not None:
            guardrail_response = await llm_service.guardrails_policy_checker.check_copilot_chat_for_policy_violations(  # noqa: E501
                context=context,
                hello_rasa_user_id=user_id,
                hello_rasa_project_id=HELLO_RASA_PROJECT_ID,
                lakera_project_id=LAKERA_COPILOT_HISTORY_GUARDRAIL_PROJECT_ID,
            )
        if guardrail_response is not None:
            blocked_or_violation_message = (
                await _handle_guardrail_violation_and_maybe_block(
                    sse=sse,
                    user_id=user_id,
                    violation_response=guardrail_response,
                )
            )

            # Send signature for the guardrail response
            if envelope := await create_signature_envelope_for_text(
                req=req,
                text=blocked_or_violation_message.content,
                category=blocked_or_violation_message.response_category,
            ):
                await sse.send(envelope.format())

            return

        # 6. Get the original response stream from copilot and handle it with the
        #    copilot response handler
        start_timestamp = time.perf_counter()
        copilot_client = llm_service.instantiate_copilot()
        (original_stream, generation_context) = await copilot_client.generate_response(
            context
        )

        copilot_response_handler = llm_service.instantiate_handler(
            COPILOT_HANDLER_ROLLING_BUFFER_SIZE
        )
        intercepted_stream = copilot_response_handler.handle_response(original_stream)

        # 7. Stream the intercepted response
        async for token in intercepted_stream:
            await sse.send(token.to_sse_event().format())

        # 8. Offload telemetry logging to a background task
        request.app.add_task(
            asyncio.to_thread(
                telemetry.log_copilot_from_handler,
                handler=copilot_response_handler,
                used_documents=generation_context.relevant_documents,
                latency_ms=int((time.perf_counter() - start_timestamp) * 1000),
                system_message=generation_context.system_message,
                chat_history=generation_context.chat_history,
                last_user_message=(
                    req.last_message.get_flattened_text_content()
                    if (req.last_message and req.last_message.role == ROLE_USER)
                    else None
                ),
                tracker_event_attachments=generation_context.tracker_event_attachments,
                **copilot_client.usage_statistics.model_dump(),
            )
        )

        # 9. Once the stream is over, extract and send references
        #    if any documents were used
        if generation_context.relevant_documents:
            reference_section = copilot_response_handler.extract_references(
                generation_context.relevant_documents
            )
            await sse.send(reference_section.to_sse_event().format())

        # 10. Sign the next history
        if envelope := await create_signature_envelope_for_handler(
            req, copilot_response_handler
        ):
            await sse.send(envelope.format())

    except CopilotStreamError as e:
        capture_exception_with_context(
            e,
            "bot_builder_service.copilot.generation_error",
            extra={"session_id": req.session_id},
            tags={"endpoint": "/api/copilot"},
        )
        await sse.send(
            ServerSentEvent(
                event="error",
                data={"error": str(e)},
            ).format()
        )

    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.copilot.unexpected_error",
            extra={"session_id": req.session_id if "req" in locals() else None},
            tags={"endpoint": "/api/copilot"},
        )
        await sse.send(
            ServerSentEvent(
                event="error",
                data={"error": str(exc)},
            ).format()
        )

    finally:
        await sse.eof()


@bp.route("/copilot/internal_message_templates/<template_name>", methods=["GET"])
@openapi.summary("Get templated response for copilot internal message")
@openapi.description(
    "Returns the templated response text for a given template name from the "
    "copilot internal message formatter. This endpoint provides access to the "
    "predefined templates used for formatting internal system messages."
)
@openapi.tag("copilot")
@openapi.parameter(
    "template_name",
    str,
    location="path",
    description=(
        "The template name to get the template for."
        "(e.g., 'training_error_log_analysis', 'e2e_testing_error_log_analysis')",
    ),
)
@openapi.response(
    200,
    {"application/json": {"template": str, "template_name": str}},
    description="Successfully retrieved the template for the given template name",
    example={
        "template": "The assistant training failed. Your task is ...",
        "template_name": "training_error_log_analysis",
    },
)
@openapi.response(
    404,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Template not found for the given template name",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def get_copilot_internal_message_template(
    request: Request, template_name: str
) -> HTTPResponse:
    """Get templated response for copilot internal message formatter."""
    try:
        # Try to get the template for the given template name
        template = llm_service.copilot_internal_message_templates.get(template_name)
        structlogger.info(
            "bot_builder_service.get_copilot_internal_message_template.template_found",
            template_name=template_name,
            template=template,
        )

        if template is None:
            structlogger.warning(
                "bot_builder_service.get_copilot_internal_message_template.not_found",
                template_name=template_name,
            )
            return response.json(
                ApiErrorResponse(
                    error="Template not found", details={"template_name": template_name}
                ).model_dump(),
                status=404,
            )

        return response.json({"template": template, "template_name": template_name})

    except Exception as e:
        structlogger.error(
            "bot_builder_service.get_copilot_internal_message_template.error",
            error=str(e),
            template_name=template_name,
        )
        return response.json(
            ApiErrorResponse(
                error="Internal server error", details={"template_name": template_name}
            ).model_dump(),
            status=500,
        )


async def current_tracker_from_input_channel(
    app: Any, session_id: str
) -> Optional[DialogueStateTracker]:
    """Generate chat bot context from current conversation."""
    if app.ctx.agent and session_id:
        return await app.ctx.agent.tracker_store.retrieve(session_id)
    else:
        return None


async def _get_copilot_block_scope(user_id: str) -> Optional[BlockScope]:
    """Return the guardrail block scope for Copilot if blocked.

    Args:
        user_id: User identifier.

    Returns:
        'user' or 'project' when blocked, otherwise None.
    """
    if not GUARDRAILS_ENABLE_BLOCKING:
        return None

    return await guardrails_store.check_block_scope(user_id)


async def _handle_guardrail_violation_and_maybe_block(
    sse: Any,
    violation_response: GeneratedContent,
    user_id: str,
) -> GeneratedContent:
    """Record a violation, apply block if threshold crossed, and respond.

    Args:
        sse: Active SSE stream.
        violation_response: The default violation warning response.
        user_id: User identifier.

    Returns:
        The GeneratedContent message that was sent to the client.
    """
    if not GUARDRAILS_ENABLE_BLOCKING:
        await sse.send(violation_response.to_sse_event().format())
        return violation_response

    result = await guardrails_store.record_violation(user_id)

    if result.user_blocked_now:
        message = CopilotResponseHandler.respond_to_guardrail_blocked(BLOCK_SCOPE_USER)
    elif result.project_blocked_now:
        message = CopilotResponseHandler.respond_to_guardrail_blocked(
            BLOCK_SCOPE_PROJECT
        )
    else:
        message = violation_response

    await sse.send(message.to_sse_event().format())
    return message



from __future__ import annotations

from typing import Any, Dict, List, Optional, Text

import structlog

from rasa.core.constants import (
    FLOW_POLICY_PRIORITY,
    POLICY_MAX_HISTORY,
    POLICY_PRIORITY,
)
from rasa.core.featurizers.tracker_featurizers import TrackerFeaturizer
from rasa.core.policies.flows import flow_executor
from rasa.core.policies.flows.flow_exceptions import FlowCircuitBreakerTrippedException
from rasa.core.policies.policy import Policy, PolicyPrediction
from rasa.dialogue_understanding.patterns.internal_error import (
    InternalErrorPatternFlowStackFrame,
)
from rasa.dialogue_understanding.stack.frames import (
    BaseFlowStackFrame,
    DialogueStackFrame,
)
from rasa.dialogue_understanding.stack.utils import (
    end_top_user_flow,
)
from rasa.engine.graph import ExecutionContext
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.core.domain import Domain
from rasa.shared.core.events import Event
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.generator import TrackerWithCachedStates
from rasa.shared.core.trackers import (
    DialogueStateTracker,
)

structlogger = structlog.get_logger()


@DefaultV1Recipe.register(
    DefaultV1Recipe.ComponentType.POLICY_WITHOUT_END_TO_END_SUPPORT, is_trainable=False
)
class FlowPolicy(Policy):
    """A policy which handles the flow of the conversation based on flows.

    Flows are loaded from files during training. During prediction,
    the flows are applied.
    """

    @staticmethod
    def does_support_stack_frame(frame: DialogueStackFrame) -> bool:
        """Checks if the policy supports the topmost frame on the dialogue stack.

        If `False` is returned, the policy will abstain from making a prediction.

        Args:
            frame: The frame to check.

        Returns:
        `True` if the policy supports the frame, `False` otherwise.
        """
        return isinstance(frame, BaseFlowStackFrame)

    @staticmethod
    def get_default_config() -> Dict[Text, Any]:
        """Returns the default config (see parent class for full docstring)."""
        # please make sure to update the docs when changing a default parameter
        return {
            POLICY_PRIORITY: FLOW_POLICY_PRIORITY,
            POLICY_MAX_HISTORY: None,
        }

    def __init__(
        self,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        featurizer: Optional[TrackerFeaturizer] = None,
    ) -> None:
        """Constructs a new Policy object."""
        super().__init__(config, model_storage, resource, execution_context, featurizer)

        self.max_history = self.config.get(POLICY_MAX_HISTORY)
        self.resource = resource

    def train(
        self,
        training_trackers: List[TrackerWithCachedStates],
        domain: Domain,
        **kwargs: Any,
    ) -> Resource:
        """Trains a policy.

        Args:
            training_trackers: The story and rules trackers from the training data.
            domain: The model's domain.
            **kwargs: Depending on the specified `needs` section and the resulting
                graph structure the policy can use different input to train itself.

        Returns:
            A policy must return its resource locator so that potential children nodes
            can load the policy from the resource.
        """
        return self.resource

    async def predict_action_probabilities(
        self,
        tracker: DialogueStateTracker,
        domain: Domain,
        rule_only_data: Optional[Dict[Text, Any]] = None,
        flows: Optional[FlowsList] = None,
        **kwargs: Any,
    ) -> PolicyPrediction:
        """Predicts the next action the bot should take after seeing the tracker.

        Args:
            tracker: The tracker containing the conversation history up to now.
            domain: The model's domain.
            rule_only_data: Slots and loops which are specific to rules and hence
                should be ignored by this policy.
            flows: The flows to use.
            **kwargs: Depending on the specified `needs` section and the resulting
                graph structure the policy can use different input to make predictions.

        Returns:
             The prediction.
        """
        if not self.supports_current_stack_frame(
            tracker
        ) or self.should_abstain_in_coexistence(tracker, True):
            # if the policy doesn't support the current stack frame, we'll abstain
            return self._prediction(self._default_predictions(domain))

        flows = flows or FlowsList(underlying_flows=[])

        # create executor and predict next action
        try:
            prediction = await flow_executor.advance_flows(
                tracker, domain.action_names_or_texts, flows, domain.slots
            )
            return self._create_prediction_result(
                prediction.action_name,
                domain,
                prediction.score,
                prediction.events,
                prediction.metadata,
            )
        except FlowCircuitBreakerTrippedException as e:
            structlogger.error(
                "flow.circuit_breaker",
                number_of_steps_taken=e.number_of_steps_taken,
                event_info=(
                    "The flow circuit breaker tripped. "
                    "There appears to be an infinite loop in the flows."
                ),
                error=str(e),
            )
            # end the current flow and start the internal error flow
            updated_stack = tracker.stack
            updated_stack = end_top_user_flow(updated_stack)
            updated_stack.push(InternalErrorPatternFlowStackFrame())
            # we retry, with the internal error frame on the stack
            events = tracker.create_stack_updated_events(updated_stack)
            tracker.update_with_events(events)
            prediction = await flow_executor.advance_flows(
                tracker, domain.action_names_or_texts, flows, domain.slots
            )
            collected_events = events + (prediction.events or [])
            return self._create_prediction_result(
                prediction.action_name,
                domain,
                prediction.score,
                collected_events,
                prediction.metadata,
            )

    def _create_prediction_result(
        self,
        action_name: Optional[Text],
        domain: Domain,
        score: float = 1.0,
        events: Optional[List[Event]] = None,
        action_metadata: Optional[Dict[Text, Any]] = None,
    ) -> PolicyPrediction:
        """Creates a prediction result.

        Args:
            action_name: The name of the predicted action.
            domain: The model's domain.
            score: The score of the predicted action.
            events: The events to return.
            action_metadata: The metadata of the predicted action.

        Returns:
            The prediction result where the score is used for one hot encoding.
        """
        result = self._default_predictions(domain)
        if action_name:
            result[domain.index_for_action(action_name)] = score
        return self._prediction(
            result, optional_events=events, action_metadata=action_metadata
        )



from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import structlog

from rasa.dialogue_understanding.patterns.validate_slot import (
    ValidateSlotPatternFlowStackFrame,
)
from rasa.dialogue_understanding.stack.dialogue_stack import DialogueStack
from rasa.dialogue_understanding.stack.frames.dialogue_stack_frame import (
    DialogueStackFrame,
)
from rasa.dialogue_understanding.stack.frames.flow_stack_frame import (
    AgentStackFrame,
    FlowStackFrameType,
    UserFlowStackFrame,
)
from rasa.dialogue_understanding.stack.frames.pattern_frame import PatternFlowStackFrame
from rasa.shared.constants import ACTION_ASK_PREFIX, UTTER_ASK_PREFIX
from rasa.shared.core.events import (
    AgentResumed,
    Event,
    FlowCompleted,
    FlowResumed,
    SlotSet,
)
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.slots import Slot
from rasa.shared.core.trackers import DialogueStateTracker

if TYPE_CHECKING:
    from rasa.dialogue_understanding.commands import StartFlowCommand

structlogger = structlog.get_logger()


def start_flow_by_name(
    flow_name: str, flows: FlowsList
) -> Optional["StartFlowCommand"]:
    from rasa.dialogue_understanding.commands import StartFlowCommand

    if flow_name in flows.user_flow_ids:
        return StartFlowCommand(flow=flow_name)
    else:
        structlogger.debug(
            "command_parser.start_flow_by_name.invalid_flow_id", flow=flow_name
        )
        return None


def extract_cleaned_options(options_str: str) -> List[str]:
    """Extract and clean options from a string."""
    delimiters = [",", " "]

    for delimiter in delimiters:
        options_str = options_str.replace(delimiter, " ")

    return sorted(
        opt.strip().strip('"').strip("'") for opt in options_str.split() if opt.strip()
    )


def is_none_value(value: str) -> bool:
    """Check if the value is a none value."""
    if not value:
        return True
    return value in {
        "[missing information]",
        "[missing]",
        "None",
        "undefined",
        "null",
    }


def clean_extracted_value(value: str) -> str:
    """Clean up the extracted value from the llm."""
    # replace any combination of single quotes, double quotes, and spaces
    # from the beginning and end of the string
    return value.strip("'\" ")


def get_nullable_slot_value(slot_value: str) -> Union[str, None]:
    """Get the slot value or None if the value is a none value.

    Args:
        slot_value: the value to coerce

    Returns:
        The slot value or None if the value is a none value.
    """
    return slot_value if not is_none_value(slot_value) else None


def initialize_pattern_validate_slot(
    slot: Slot,
) -> Optional[ValidateSlotPatternFlowStackFrame]:
    """Initialize the pattern to validate a slot value."""
    if not slot.requires_validation():
        return None

    validation = slot.validation
    slot_name = slot.name
    return ValidateSlotPatternFlowStackFrame(
        validate=slot_name,
        refill_utter=validation.refill_utter or f"{UTTER_ASK_PREFIX}{slot_name}",  # type: ignore[union-attr]
        refill_action=f"{ACTION_ASK_PREFIX}{slot_name}",
        rejections=validation.rejections,  # type: ignore[union-attr]
    )


def create_validate_frames_from_slot_set_events(
    tracker: DialogueStateTracker,
    events: List[Event],
    validate_frames: List[ValidateSlotPatternFlowStackFrame] = [],
    should_break: bool = False,
) -> Tuple[DialogueStateTracker, List[ValidateSlotPatternFlowStackFrame]]:
    """Process SlotSet events and create validation frames.

    Args:
        tracker: The dialogue state tracker.
        events: List of events to process.
        validate_frames: List to collect validation frames.
        should_break:  whether to break after the first non-SlotSet event.
            if True, break out of the event loop as soon as the first non-SlotSet
            event is encountered.
            if False, continue processing the events until the end.

    Returns:
        Tuple of (updated tracker, list of validation frames).
    """
    for event in events:
        if not isinstance(event, SlotSet):
            if should_break:
                # we want to only process the most recent SlotSet events
                # so we break once we encounter a different event
                break
            continue

        slot = tracker.slots.get(event.key)
        frame = initialize_pattern_validate_slot(slot)

        if frame:
            validate_frames.append(frame)

    return tracker, validate_frames


def find_default_flows_collecting_slot(
    slot_name: str, all_flows: FlowsList
) -> List[str]:
    """Find default flows that have collect steps matching the specified slot name.

    Args:
        slot_name: The name of the slot to search for.
        all_flows: All flows in the assistant.

    Returns:
        List of flow IDs for default flows that collect the specified slot
        without asking before filling.
    """
    return [
        flow.id
        for flow in all_flows.underlying_flows
        if flow.is_rasa_default_flow
        and any(
            step.collect == slot_name and not step.ask_before_filling
            for step in flow.get_collect_steps()
        )
    ]


def resume_flow(
    flow_to_resume: str,
    tracker: DialogueStateTracker,
    stack: DialogueStack,
) -> List[Event]:
    """Resumes a flow by reordering frames."""
    applied_events: List[Event] = []

    # Resume existing flow by reordering frames
    frames_to_resume, user_frame_to_resume = collect_frames_to_resume(
        stack, flow_to_resume
    )

    # if the flow is not on the stack, do nothing
    # this should not happen, but just in case
    if user_frame_to_resume is None:
        structlogger.error(
            "resume_flow.no_user_frame_to_resume",
            flow_to_resume=flow_to_resume,
        )
        return []

    # move the frames to the top of the stack, e.g. reorder the frames
    # on the stack
    stack.move_frames_to_top(frames_to_resume)

    # create agent resumed events if the agent frame is now on top of the stack
    agent_stack_frame = next(
        (frame for frame in frames_to_resume if isinstance(frame, AgentStackFrame)),
        None,
    )
    if agent_stack_frame:
        agent_id = agent_stack_frame.agent_id
        applied_events.append(AgentResumed(agent_id, agent_stack_frame.flow_id))

    # Create flow interruption and resumption events
    applied_events.extend(
        [
            # the flow, which was on the stack, is resumed
            FlowResumed(user_frame_to_resume.flow_id, user_frame_to_resume.step_id),
        ]
    )

    return applied_events + tracker.create_stack_updated_events(stack)


def collect_frames_to_resume(
    stack: DialogueStack,
    target_flow_id: str,  # pyright: ignore[reportUndefinedVariable]
) -> Tuple[List[DialogueStackFrame], Optional[UserFlowStackFrame]]:
    """Collect frames that need to be resumed for the target flow.

    Args:
        stack: The stack to collect frames from.
        target_flow_id: The ID of the flow to resume.

    Returns:
        A tuple containing (frames_to_resume, frame_to_resume).
    """
    frames_to_resume: List[DialogueStackFrame] = []
    frame_found = False
    frame_to_resume = None

    for frame in stack.frames:
        if isinstance(frame, UserFlowStackFrame) and (
            frame.frame_type == FlowStackFrameType.REGULAR
            or frame.frame_type == FlowStackFrameType.INTERRUPT
        ):
            if frame.flow_id == target_flow_id:
                frames_to_resume.append(frame)
                frame_to_resume = frame
                frame_found = True
                continue
            elif frame_found:
                break

        if frame_found:
            frames_to_resume.append(frame)

    return list(frames_to_resume), frame_to_resume


def remove_pattern_continue_interrupted_frames(
    stack: DialogueStack,
) -> Tuple[DialogueStack, List[FlowCompleted]]:
    """Remove pattern_continue_interrupted frames from the stack and return events.

    Returns:
        A tuple containing (updated_stack, flow_completed_events)
    """
    from rasa.dialogue_understanding.stack.utils import (
        is_continue_interrupted_flow_active,
    )

    if not is_continue_interrupted_flow_active(stack):
        return stack, []

    events = []
    # remove pattern_continue_interrupted from the stack
    top_frame = stack.top()
    while isinstance(top_frame, PatternFlowStackFrame):
        # Create FlowCompleted event for the pattern frame being removed
        events.append(FlowCompleted(top_frame.flow_id, top_frame.step_id))

        # If the top frame is a pattern frame, we need to remove it
        # before continuing with the active user flow frame.
        # This prevents the pattern frame
        # from being left on the stack when the flow is started
        # which would prevent pattern_completed to be triggered
        # once the user flow is completed.
        stack.pop()
        top_frame = stack.top()

    return stack, events



import datetime as dt
import os
import uuid
from typing import Any, Iterable, Optional, Sequence

import structlog

from rasa import telemetry
from rasa.builder.copilot.constants import COPILOT_SEGMENT_WRITE_KEY_ENV_VAR
from rasa.builder.copilot.copilot_response_handler import CopilotResponseHandler
from rasa.builder.copilot.models import EventContent
from rasa.builder.document_retrieval.models import Document
from rasa.telemetry import (
    SEGMENT_TRACK_ENDPOINT,
    segment_request_payload,
    send_segment_request,
    with_default_context_fields,
)

structlogger = structlog.get_logger()

COPILOT_USER_MESSAGE_EVENT = "copilot_user_message"
COPILOT_BOT_MESSAGE_EVENT = "copilot_bot_message"

COPILOT_SEGMENT_WRITE_KEY = os.getenv(COPILOT_SEGMENT_WRITE_KEY_ENV_VAR)
if _SEGMENT_ON := bool(COPILOT_SEGMENT_WRITE_KEY):
    structlogger.info("builder.telemetry.enabled")
else:
    structlogger.warning("builder.telemetry.disabled")


def _track(event: str, user_id: str, properties: dict) -> None:
    """Track an event with Segment.

    Args:
        event: The name of the event to track.
        user_id: The ID of the user associated with the event.
        properties: Additional properties to include with the event.

    Raises:
        Exception: If tracking fails, an exception is logged.
    """
    if not _SEGMENT_ON or not telemetry.is_telemetry_enabled():
        structlogger.debug("builder.telemetry._track.disabled")
        return
    structlogger.debug("builder.telemetry._track.enabled")

    try:
        payload = segment_request_payload(
            user_id, event, properties, context=with_default_context_fields()
        )
        structlogger.debug("builder.telemetry._track.sending", payload=payload)

        send_segment_request(SEGMENT_TRACK_ENDPOINT, payload, COPILOT_SEGMENT_WRITE_KEY)
    except Exception as e:  # skipcq:PYL-W0703
        structlogger.warning("builder.telemetry.track_failed", error=str(e))


class CopilotTelemetry:
    def __init__(
        self,
        *,
        project_id: str,
        user_id: str,
    ) -> None:
        """Initialize Telemetry instance."""
        self._project_id = project_id
        self._user_id = user_id
        # TODO Load prompt version
        self._prompt_version = "1"

    def log_user_turn(self, text: str) -> None:
        """Track a user message in the conversation.

        Args:
            text: The text of the user message.
        """
        structlogger.debug("builder.telemetry.log_user_turn", text=text)
        _track(
            COPILOT_USER_MESSAGE_EVENT,
            self._user_id,
            {
                "project_id": self._project_id,
                "message_id": uuid.uuid4().hex,
                "text": text,
                "timestamp": dt.datetime.utcnow().isoformat(),
            },
        )

    def log_copilot_turn(
        self,
        *,
        text: str,
        source_urls: Sequence[str],
        flags: Iterable[str],
        latency_ms: int,
        model: str,
        input_tokens: Optional[int] = None,
        output_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        system_message: Optional[dict[str, Any]] = None,
        chat_history: Optional[list[dict[str, Any]]] = None,
        last_user_message: Optional[str] = None,
        tracker_event_attachments: Optional[list[dict[str, Any]]] = None,
    ) -> None:
        """Track a copilot message in the conversation.

        Args:
            text: The text of the copilot message.
            source_urls: URLs of the sources used to generate the response.
            flags: Flags indicating special conditions or features.
            latency_ms: End-to-end Copilot latency to produce this response.
            model: The model used to generate the response.
            input_tokens: Number of input tokens used (optional).
            output_tokens: Number of output tokens generated (optional).
            total_tokens: Total number of tokens used (input + output) (optional).
            system_message: The system message used (optional).
            chat_history: The chat history messages used (optional).
            last_user_message: The last user message used (optional).
            tracker_event_attachments: The tracker event attachments used (optional).
        """
        structlogger.debug("builder.telemetry.log_copilot_turn", text=text)

        # FIXME: Temporarily remove the system_message from telemetry payload.
        # Reason: It often exceeds Segment payload size limits, causing the request
        # to be rejected and the event to be absent in Segment. Instead, temporarily
        # log the system_message so it's visible in Grafana.
        telemetry_data = {
            "project_id": self._project_id,
            "message_id": uuid.uuid4().hex,
            "text": text,
            "prompt_version": self._prompt_version,
            "source_urls": list(source_urls),
            "flags": list(flags),
            "latency_ms": latency_ms,
            "model": model,
            "input_tokens": input_tokens,
            "output_tokens": output_tokens,
            "total_tokens": total_tokens,
            "chat_history": chat_history,
            "last_user_message": last_user_message,
            "tracker_event_attachments": tracker_event_attachments,
            "timestamp": dt.datetime.utcnow().isoformat(),
        }

        # Log all telemetry data plus system_message for debugging
        log_data = telemetry_data.copy()
        if system_message:
            log_data["system_message"] = system_message

        structlogger.info("builder.telemetry.copilot_turn", **log_data)

        _track(
            COPILOT_BOT_MESSAGE_EVENT,
            self._user_id,
            telemetry_data,
        )

    @staticmethod
    def _extract_flags(handler: CopilotResponseHandler) -> list[str]:
        """Extract flags from the response handler.

        Args:
            handler: The response handler containing generated responses.

        Returns:
            A list of flags indicating special conditions or features.
        """
        flags = {r.response_category.value for r in handler.generated_responses}
        return sorted(flags)

    @staticmethod
    def _full_text(handler: CopilotResponseHandler) -> str:
        """Extract full text from the response handler.

        Args:
            handler: The response handler containing generated responses.

        Returns:
            The concatenated content of all generated responses.
        """
        return "".join(
            response.content
            for response in handler.generated_responses
            if getattr(response, "content", None)
        )

    def log_copilot_from_handler(
        self,
        *,
        handler: CopilotResponseHandler,
        used_documents: list[Document],
        latency_ms: int,
        model: str,
        prompt_tokens: int,
        completion_tokens: int,
        total_tokens: int,
        system_message: dict[str, Any],
        chat_history: list[dict[str, Any]],
        last_user_message: Optional[str],
        tracker_event_attachments: list[EventContent],
    ) -> None:
        """Log a copilot message from the response handler.

        Args:
            handler: The response handler containing generated responses.
            used_documents: List of documents used as supporting evidence.
            latency_ms: End-to-end Copilot latency to produce this response.
            model: The model used to generate the response.
            prompt_tokens: Number of input tokens used.
            completion_tokens: Number of output tokens generated.
            total_tokens: Total number of tokens used (input + output).
            system_message: The system message used.
            chat_history: The chat history messages used.
            last_user_message: The last user message used.
        """
        structlogger.debug("builder.telemetry.log_copilot_from_handler")
        text = self._full_text(handler)
        self.log_copilot_turn(
            text=text,
            source_urls=[d.url for d in used_documents if d.url],
            flags=self._extract_flags(handler),
            latency_ms=latency_ms,
            model=model,
            input_tokens=prompt_tokens,
            output_tokens=completion_tokens,
            total_tokens=total_tokens,
            system_message=system_message,
            chat_history=chat_history,
            last_user_message=last_user_message,
            tracker_event_attachments=[
                attachment.model_dump() for attachment in tracker_event_attachments
            ],
        )



import json
import time
from collections import deque
from typing import Deque, Optional, Text

import structlog
from pydantic import ValidationError

from rasa.core.iam_credentials_providers.credentials_provider_protocol import (
    SupportedServiceType,
)
from rasa.core.lock import Ticket, TicketLock
from rasa.core.lock_store import (
    DEFAULT_SOCKET_TIMEOUT_IN_SECONDS,
    LOCK_LIFETIME,
    LockError,
    LockStore,
)
from rasa.core.redis_connection_factory import (
    DeploymentMode,
    RedisConfig,
    RedisConnectionFactory,
)
from rasa.shared.exceptions import RasaException
from rasa.utils.endpoints import EndpointConfig

DEFAULT_REDIS_DB = 1

DEFAULT_PORT = 6379

DEFAULT_HOSTNAME = "localhost"

structlogger = structlog.getLogger(__name__)

LAST_ISSUED_TICKET_NUMBER_SUFFIX = "last_issued_ticket_number"

DEFAULT_CONCURRENT_REDIS_LOCK_STORE_KEY_PREFIX = "concurrent_lock:"


class ConcurrentTicketLock(TicketLock):
    def concurrent_issue_ticket(self, lifetime: float, ticket_number: int) -> int:
        """Issue a new ticket and return its number.

        Args:
            lifetime: The length of time after which the ticket expires.
            ticket_number: Integer representing the ticket number.

        Returns:
            ticket_number integer.

        """
        ticket = Ticket(ticket_number, time.time() + lifetime)
        self.tickets.append(ticket)

        return ticket_number


class ConcurrentRedisLockStore(LockStore):
    """Concurrent implementation of a Redis store for ticket locks."""

    def __init__(
        self,
        endpoint_config: EndpointConfig,
    ) -> None:
        """Create a lock store which uses Redis for persistence.

        Args:
            endpoint_config: The endpointConfig defined the lock store.

            Can contain the following properties:
            host - The host of the redis server.
            port - The port of the redis server.
            db - The name of the database within Redis which should be used by Rasa
                Open Source.
            username - The username which should be used for authentication with the
                Redis database.
            password - The password which should be used for authentication with the
                Redis database.
            use_ssl - `True` if SSL should be used for the connection to Redis.
            ssl_certfile - Path to the SSL certificate file.
            ssl_keyfile - Path to the SSL private key file.
            ssl_ca_certs - Path to the SSL CA certificate file.
            key_prefix - prefix to prepend to all keys used by the lock store. Must be
                alphanumeric.
            socket_timeout - Timeout in seconds after which an exception will be raised
                in case Redis doesn't respond within `socket_timeout` seconds.
            deployment_mode - Redis deployment mode: standard, cluster, or sentinel.
            endpoints - List of endpoints for cluster/sentinel mode in host:port format.
            sentinel_service - Sentinel service name.
        """
        host = endpoint_config.kwargs.get("host", DEFAULT_HOSTNAME)
        port = endpoint_config.kwargs.get("port", DEFAULT_PORT)
        db = endpoint_config.kwargs.get("db", DEFAULT_REDIS_DB)
        username = endpoint_config.kwargs.get("username")
        password = endpoint_config.kwargs.get("password")
        use_ssl = bool(endpoint_config.kwargs.get("use_ssl", False))
        ssl_certfile = endpoint_config.kwargs.get("ssl_certfile")
        ssl_keyfile = endpoint_config.kwargs.get("ssl_keyfile")
        ssl_ca_certs = endpoint_config.kwargs.get("ssl_ca_certs")
        key_prefix = endpoint_config.kwargs.get("key_prefix")
        socket_timeout = endpoint_config.kwargs.get(
            "socket_timeout", DEFAULT_SOCKET_TIMEOUT_IN_SECONDS
        )
        deployment_mode = endpoint_config.kwargs.get(
            "deployment_mode", DeploymentMode.STANDARD.value
        )
        endpoints = endpoint_config.kwargs.get("endpoints", [])
        sentinel_service = endpoint_config.kwargs.get("sentinel_service")

        try:
            redis_config = RedisConfig(
                host=host,
                port=port,
                service_type=SupportedServiceType.LOCK_STORE,
                db=db,
                username=username,
                password=password,
                use_ssl=use_ssl,
                ssl_certfile=ssl_certfile,
                ssl_keyfile=ssl_keyfile,
                ssl_ca_certs=ssl_ca_certs,
                socket_timeout=socket_timeout,
                deployment_mode=deployment_mode,
                endpoints=endpoints,
                sentinel_service=sentinel_service,
            )
            self.red = RedisConnectionFactory.create_connection(redis_config)
        except ValidationError as e:
            raise RasaException(f"Invalid Redis configuration: {e}")

        self.deployment_mode = deployment_mode
        self.key_prefix = DEFAULT_CONCURRENT_REDIS_LOCK_STORE_KEY_PREFIX
        if key_prefix:
            structlogger.debug(
                "concurrent_redis_lock_store._set_key_prefix.non_default_key_prefix",
                event_info=f"Setting non-default redis key prefix: '{key_prefix}'.",
            )
            self._set_key_prefix(key_prefix)

        super().__init__()

    def _set_key_prefix(self, key_prefix: Text) -> None:
        if isinstance(key_prefix, str) and key_prefix.isalnum():
            self.key_prefix = (
                key_prefix + ":" + DEFAULT_CONCURRENT_REDIS_LOCK_STORE_KEY_PREFIX
            )
        else:
            structlogger.warning(
                "concurrent_redis_lock_store._set_key_prefix.default_instead_of_invalid_key_prefix",
                event_info=(
                    f"Omitting provided non-alphanumeric "
                    f"redis key prefix: '{key_prefix}'. "
                    f"Using default '{self.key_prefix}' instead."
                ),
            )

    def _scan_cluster_keys(self, pattern: Text) -> list:
        """Scan keys in cluster mode with proper cursor handling."""
        keys = []
        cursor = 0

        while True:
            try:
                cursor, batch_keys = self.red.scan(cursor, match=pattern, count=100)
                keys.extend(batch_keys)

                if isinstance(cursor, dict):
                    # cursor is a dict mapping each node to its scan position. e.g
                    # {'127.0.0.1:7000': 0, '127.0.0.1:7001': 5, '127.0.0.1:7002': 0}
                    # A cursor value of 0 means that node has finished scanning
                    # When all nodes show 0, the entire cluster scan is complete
                    if all(v == 0 for v in cursor.values()):
                        break
                else:
                    # if scan is complete
                    if cursor == 0:
                        break

            except Exception as e:
                structlogger.warning(
                    "concurrent_redis_lock_store._get_keys_by_pattern.scan_interrupted",
                    event_info=f"SCAN interrupted in cluster mode: {e}. "
                    f"Returning {len(keys)} keys found so far.",
                )
                break

        return keys

    def _get_keys_by_pattern(self, pattern: Text) -> list:
        """Get keys by pattern, using SCAN for cluster mode and KEYS for others."""
        if self.deployment_mode == DeploymentMode.CLUSTER.value:
            return self._scan_cluster_keys(pattern)
        else:
            return self.red.keys(pattern)

    def issue_ticket(
        self, conversation_id: Text, lock_lifetime: float = LOCK_LIFETIME
    ) -> int:
        """Issue new ticket for a conversation.

        It's configured with `lock_lifetime` and associated with `conversation_id`.
        Creates a new lock if none is found.
        """
        structlogger.debug(
            "concurrent_redis_lock_store.issue_ticket",
            event_info=f"Issuing ticket for conversation '{conversation_id}'.",
        )
        try:
            lock = self.get_or_create_lock(conversation_id)
            lock.remove_expired_tickets()
            ticket_number = self.increment_ticket_number(lock)
            ticket = lock.concurrent_issue_ticket(lock_lifetime, ticket_number)  # type: ignore[attr-defined]
            self.save_lock(lock)

            return ticket
        except Exception as e:
            raise LockError(f"Error while acquiring lock. Error:\n{e}")

    def get_lock(self, conversation_id: Text) -> Optional[ConcurrentTicketLock]:
        """Retrieves lock (see parent docstring for more information)."""
        tickets: Deque[Ticket] = deque()

        pattern = self.key_prefix + conversation_id + ":" + "[0-9]*"
        redis_keys = self._get_keys_by_pattern(pattern)

        for key in redis_keys:
            serialised_ticket = self.red.get(key)
            if serialised_ticket:
                # Handle bytes to string conversion for JSON parsing
                if isinstance(serialised_ticket, bytes):
                    serialised_ticket = serialised_ticket.decode("utf-8")
                ticket = Ticket.from_dict(json.loads(serialised_ticket))
                tickets.appendleft(ticket)

        tickets = deque(sorted(tickets, key=lambda x: x.number))

        return ConcurrentTicketLock(conversation_id, tickets)

    def delete_lock(self, conversation_id: Text) -> None:
        """Deletes lock for conversation ID."""
        pattern = self.key_prefix + conversation_id + ":*"
        redis_keys = self._get_keys_by_pattern(pattern)

        if not redis_keys:
            structlogger.debug(
                "concurrent_redis_lock_store.delete_lock_key_not_found",
                event_info=(
                    f"The lock store does not contain any key-value "
                    f"items for conversation '{conversation_id}'."
                ),
            )
            return None

        deletion_successful = self.red.delete(*redis_keys)
        if deletion_successful == 0:
            self._log_deletion(conversation_id, False)
        else:
            self._log_deletion(conversation_id, True)

    def save_lock(self, lock: TicketLock) -> None:
        """Commit individual tickets and last issued ticket number to storage."""
        last_issued_ticket = lock.tickets[-1]
        serialised_ticket = last_issued_ticket.dumps()
        key = (
            self.key_prefix
            + lock.conversation_id
            + ":"
            + str(last_issued_ticket.number)
        )
        self.red.set(
            name=key, value=serialised_ticket, ex=int(last_issued_ticket.expires)
        )

    def increment_ticket_number(self, lock: TicketLock) -> int:
        """Uses Redis atomic transaction to increment ticket number."""
        last_issued_key = (
            self.key_prefix
            + lock.conversation_id
            + ":"
            + LAST_ISSUED_TICKET_NUMBER_SUFFIX
        )

        return self.red.incr(name=last_issued_key)

    def finish_serving(self, conversation_id: Text, ticket_number: int) -> None:
        """Finish serving ticket with `ticket_number` for `conversation_id`.

        Removes ticket from storage.
        """
        ticket_key = self.key_prefix + conversation_id + ":" + str(ticket_number)
        self.red.delete(ticket_key)



from __future__ import annotations

import argparse
import functools
import sys
from typing import TYPE_CHECKING, List, Optional, Text, Union

import pluggy

from rasa.cli import SubParsersAction

if TYPE_CHECKING:
    from rasa.core.brokers.broker import EventBroker
    from rasa.core.tracker_stores.tracker_store import TrackerStore
    from rasa.shared.core.domain import Domain
    from rasa.shared.core.trackers import DialogueStateTracker
    from rasa.utils.endpoints import EndpointConfig


hookspec = pluggy.HookspecMarker("rasa")


@functools.lru_cache(maxsize=2)
def plugin_manager() -> pluggy.PluginManager:
    """Initialises a plugin manager which registers hook implementations."""
    _plugin_manager = pluggy.PluginManager("rasa")
    _plugin_manager.add_hookspecs(sys.modules["rasa.plugin"])
    init_hooks(_plugin_manager)

    return _plugin_manager


def init_hooks(manager: pluggy.PluginManager) -> None:
    """Initialise hooks into rasa."""
    from rasa import hooks

    manager.register(hooks)


@hookspec
def refine_cli(
    subparsers: SubParsersAction,
    parent_parsers: List[argparse.ArgumentParser],
) -> None:
    """Customizable hook for adding CLI commands."""


@hookspec
def configure_commandline(cmdline_arguments: argparse.Namespace) -> Optional[Text]:
    """Hook specification for configuring plugin CLI."""


@hookspec
def init_telemetry(endpoints_file: Optional[Text]) -> None:
    """Hook specification for initialising plugin telemetry."""


@hookspec
def init_managers(endpoints_file: Optional[Text]) -> None:
    """Hook specification for initialising managers."""


@hookspec(firstresult=True)
def create_tracker_store(  # type: ignore[empty-body]
    endpoint_config: Union["TrackerStore", "EndpointConfig"],
    domain: "Domain",
    event_broker: Optional["EventBroker"],
) -> "TrackerStore":
    """Hook specification for wrapping with AuthRetryTrackerStore."""


@hookspec
def after_server_stop() -> None:
    """Hook specification for stopping the server.

    Use this hook to de-initialize any resources that require explicit cleanup like,
    thread shutdown, closing connections, etc.
    """


@hookspec
def after_new_user_message(tracker: "DialogueStateTracker") -> None:
    """Hook specification for after a new user message is received."""


@hookspec
def after_action_executed(tracker: "DialogueStateTracker") -> None:
    """Hook specification for after an action is executed."""



from typing import Any, Dict, Literal, Optional, Text

import structlog

from rasa.core.config.configuration import Configuration
from rasa.dialogue_understanding.commands.command_syntax_manager import (
    CommandSyntaxVersion,
)
from rasa.dialogue_understanding.generator.constants import (
    DEFAULT_OPENAI_MAX_GENERATED_TOKENS,
    LLM_CONFIG_KEY,
    MODEL_CONFIG_KEY,
    MODEL_NAME_CLAUDE_3_5_SONNET_20240620,
    MODEL_NAME_GPT_4O_2024_11_20,
    OPENAI_PROVIDER,
    PROVIDER_CONFIG_KEY,
    TIMEOUT_CONFIG_KEY,
)
from rasa.dialogue_understanding.generator.single_step.single_step_based_llm_command_generator import (  # noqa: E501  # noqa: E501
    SingleStepBasedLLMCommandGenerator,
)
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.constants import (
    ANTHROPIC_PROVIDER,
    AWS_BEDROCK_PROVIDER,
    AZURE_OPENAI_PROVIDER,
    MAX_COMPLETION_TOKENS_CONFIG_KEY,
    PROMPT_TEMPLATE_CONFIG_KEY,
    TEMPERATURE_CONFIG_KEY,
)
from rasa.shared.utils.llm import (
    get_default_prompt_template_based_on_model,
    get_prompt_template,
)

structlogger = structlog.get_logger()

DEFAULT_LLM_CONFIG = {
    PROVIDER_CONFIG_KEY: OPENAI_PROVIDER,
    MODEL_CONFIG_KEY: MODEL_NAME_GPT_4O_2024_11_20,
    TEMPERATURE_CONFIG_KEY: 0.0,
    MAX_COMPLETION_TOKENS_CONFIG_KEY: DEFAULT_OPENAI_MAX_GENERATED_TOKENS,
    TIMEOUT_CONFIG_KEY: 7,
}

# Non-agent defaults
DEFAULT_COMMAND_PROMPT_TEMPLATE_FILE_NAME = (
    "command_prompt_v3_gpt_4o_2024_11_20_template.jinja2"
)
FALLBACK_COMMAND_PROMPT_TEMPLATE_FILE_NAME = (
    "command_prompt_v3_gpt_4o_2024_11_20_template.jinja2"
)
MODEL_PROMPT_MAPPER = {
    f"{OPENAI_PROVIDER}/{MODEL_NAME_GPT_4O_2024_11_20}": (
        "command_prompt_v3_gpt_4o_2024_11_20_template.jinja2"
    ),
    f"{AZURE_OPENAI_PROVIDER}/{MODEL_NAME_GPT_4O_2024_11_20}": (
        "command_prompt_v3_gpt_4o_2024_11_20_template.jinja2"
    ),
    f"{AWS_BEDROCK_PROVIDER}/anthropic."
    f"{MODEL_NAME_CLAUDE_3_5_SONNET_20240620}-v1:0": (
        "command_prompt_v3_claude_3_5_sonnet_20240620_template.jinja2"
    ),
    f"{ANTHROPIC_PROVIDER}/{MODEL_NAME_CLAUDE_3_5_SONNET_20240620}": (
        "command_prompt_v3_claude_3_5_sonnet_20240620_template.jinja2"
    ),
}

# Agentic mapping and defaults (used only when agents exist)
AGENT_DEFAULT_COMMAND_PROMPT_TEMPLATE_FILE_NAME = (
    "agent_command_prompt_v3_gpt_4o_2024_11_20_template.jinja2"
)
AGENT_FALLBACK_COMMAND_PROMPT_TEMPLATE_FILE_NAME = (
    "agent_command_prompt_v3_gpt_4o_2024_11_20_template.jinja2"
)
AGENT_MODEL_PROMPT_MAPPER = {
    f"{OPENAI_PROVIDER}/{MODEL_NAME_GPT_4O_2024_11_20}": (
        "agent_command_prompt_v3_gpt_4o_2024_11_20_template.jinja2"
    ),
    f"{AZURE_OPENAI_PROVIDER}/{MODEL_NAME_GPT_4O_2024_11_20}": (
        "agent_command_prompt_v3_gpt_4o_2024_11_20_template.jinja2"
    ),
    f"{AWS_BEDROCK_PROVIDER}/anthropic."
    f"{MODEL_NAME_CLAUDE_3_5_SONNET_20240620}-v1:0": (
        "agent_command_prompt_v3_claude_3_5_sonnet_20240620_template.jinja2"
    ),
    f"{ANTHROPIC_PROVIDER}/{MODEL_NAME_CLAUDE_3_5_SONNET_20240620}": (
        "agent_command_prompt_v3_claude_3_5_sonnet_20240620_template.jinja2"
    ),
}


@DefaultV1Recipe.register(
    [
        DefaultV1Recipe.ComponentType.COMMAND_GENERATOR,
    ],
    is_trainable=True,
)
class SearchReadyLLMCommandGenerator(SingleStepBasedLLMCommandGenerator):
    """A single step LLM-based command generator."""

    def __init__(
        self,
        config: Dict[str, Any],
        model_storage: ModelStorage,
        resource: Resource,
        prompt_template: Optional[Text] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            config,
            model_storage,
            resource,
            prompt_template=prompt_template,
            **kwargs,
        )

    @staticmethod
    def get_default_llm_config() -> Dict[str, Any]:
        """Get the default LLM config for the command generator."""
        return DEFAULT_LLM_CONFIG

    @staticmethod
    def get_default_prompt_template_file_name() -> str:
        """Get the default prompt template file name for the command generator."""
        return (
            AGENT_DEFAULT_COMMAND_PROMPT_TEMPLATE_FILE_NAME
            if Configuration.get_instance().available_agents.has_agents()
            else DEFAULT_COMMAND_PROMPT_TEMPLATE_FILE_NAME
        )

    @staticmethod
    def get_fallback_prompt_template_file_name() -> str:
        """Get the fallback prompt template file name for the command generator."""
        return (
            AGENT_FALLBACK_COMMAND_PROMPT_TEMPLATE_FILE_NAME
            if Configuration.get_instance().available_agents.has_agents()
            else FALLBACK_COMMAND_PROMPT_TEMPLATE_FILE_NAME
        )

    @staticmethod
    def get_model_prompt_mapper() -> Dict[str, str]:
        """Get the model prompt mapper for the command generator."""
        return (
            AGENT_MODEL_PROMPT_MAPPER
            if Configuration.get_instance().available_agents.has_agents()
            else MODEL_PROMPT_MAPPER
        )

    @staticmethod
    def get_component_command_syntax_version() -> CommandSyntaxVersion:
        return CommandSyntaxVersion.v3

    @classmethod
    def _resolve_component_prompt_template(
        cls,
        config: Dict[str, Any],
        prompt_template: Optional[str] = None,
        log_context: Optional[Literal["init", "fingerprint_addon"]] = None,
        log_source_component: Optional[str] = "SearchReadyLLMCommandGenerator",
    ) -> Optional[str]:
        """Get the prompt template from the config or the default prompt template."""
        if prompt_template is not None:
            return prompt_template

        # Try to load the template from the given path or fallback to the default for
        # the component.
        custom_prompt_template_path = config.get(PROMPT_TEMPLATE_CONFIG_KEY)
        if custom_prompt_template_path is not None:
            custom_prompt_template = get_prompt_template(
                custom_prompt_template_path,
                None,  # Default will be based on the model
                log_source_component=log_source_component,
                log_source_method=log_context,
            )
            if custom_prompt_template is not None:
                return custom_prompt_template

        # Fallback to the default prompt template based on the model.
        default_command_prompt_template = get_default_prompt_template_based_on_model(
            llm_config=config.get(LLM_CONFIG_KEY, {}) or {},
            model_prompt_mapping=cls.get_model_prompt_mapper(),
            default_prompt_path=cls.get_default_prompt_template_file_name(),
            fallback_prompt_path=cls.get_fallback_prompt_template_file_name(),
            log_source_component=log_source_component,
            log_source_method=log_context,
        )

        return default_command_prompt_template



import argparse
import logging
from enum import Enum
from typing import List, Optional, Text, Union

from rasa.core.constants import DEFAULT_SUB_AGENTS
from rasa.core.persistor import RemoteStorageType, StorageType, parse_remote_storage
from rasa.shared.constants import (
    DEFAULT_CONFIG_PATH,
    DEFAULT_DATA_PATH,
    DEFAULT_ENDPOINTS_PATH,
    DEFAULT_MODELS_PATH,
)


def add_model_param(
    parser: Union[argparse.ArgumentParser, argparse._ActionsContainer],
    model_name: Text = "Rasa",
    add_positional_arg: bool = True,
    default: Optional[Text] = DEFAULT_MODELS_PATH,
) -> None:
    help_text = (
        "Path to a trained {} model. If a directory is specified, it will "
        "use the latest model in this directory.".format(model_name)
    )
    parser.add_argument("-m", "--model", type=str, default=default, help=help_text)
    if add_positional_arg:
        parser.add_argument(
            "model-as-positional-argument", nargs="?", type=str, help=help_text
        )


def add_stories_param(
    parser: Union[argparse.ArgumentParser, argparse._ActionsContainer],
    stories_name: Text = "training",
) -> None:
    parser.add_argument(
        "-s",
        "--stories",
        type=str,
        default=DEFAULT_DATA_PATH,
        help=f"File or folder containing your {stories_name} stories.",
    )


def add_nlu_data_param(
    parser: Union[argparse.ArgumentParser, argparse._ActionsContainer],
    help_text: Text,
    default: Optional[Text] = DEFAULT_DATA_PATH,
) -> None:
    parser.add_argument("-u", "--nlu", type=str, default=default, help=help_text)


def add_domain_param(
    parser: Union[argparse.ArgumentParser, argparse._ActionsContainer],
    default: Optional[Text] = None,
) -> None:
    parser.add_argument(
        "-d",
        "--domain",
        type=str,
        default=default,
        help="Domain specification. This can be a single YAML file, or a directory "
        "that contains several files with domain specifications in it. The content "
        "of these files will be read and merged together.",
    )


def add_config_param(
    parser: Union[argparse.ArgumentParser, argparse._ActionsContainer],
    default: Optional[Text] = DEFAULT_CONFIG_PATH,
) -> None:
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        default=default,
        help="The policy and pipeline configuration of your bot.",
    )


def add_out_param(
    parser: Union[argparse.ArgumentParser, argparse._ActionsContainer],
    help_text: Text,
    default: Optional[Text] = DEFAULT_MODELS_PATH,
    required: bool = False,
) -> None:
    parser.add_argument(
        "--out",
        type=str,
        default=default,
        help=help_text,
        # The desired behaviour is that required indicates if this argument must
        # have a value, but argparse interprets it as "must have a value
        # from user input", so we toggle it only if our default is not set
        required=required and default is None,
    )


def add_endpoint_param(
    parser: Union[argparse.ArgumentParser, argparse._ActionsContainer],
    help_text: Text,
    default: Optional[Text] = DEFAULT_ENDPOINTS_PATH,
) -> None:
    """Adds an option to an argument parser to configure endpoints path."""
    parser.add_argument("--endpoints", type=str, default=default, help=help_text)


def add_data_param(
    parser: Union[argparse.ArgumentParser, argparse._ActionsContainer],
    default: Optional[Text] = DEFAULT_DATA_PATH,
    required: bool = False,
    data_type: Text = "Rasa ",
) -> None:
    parser.add_argument(
        "--data",
        default=default,
        nargs="+",
        type=str,
        help=f"Paths to the files or directories containing {data_type} data.",
        # The desired behaviour is that required indicates if this argument must
        # have a value, but argparse interprets it as "must have a value
        # from user input", so we toggle it only if our default is not set
        required=required and default is None,
    )


def add_logging_options(parser: argparse.ArgumentParser) -> None:
    """Add options to an argument parser to configure logging levels."""
    logging_arguments = parser.add_argument_group(
        "Python Logging Options",
        "You can control level of log messages printed. "
        "In addition to these arguments, a more fine grained configuration can be "
        "achieved with environment variables. See online documentation for more info.",
    )

    # arguments for logging configuration
    logging_arguments.add_argument(
        "-v",
        "--verbose",
        help="Be verbose. Sets logging level to INFO.",
        action="store_const",
        dest="loglevel",
        const=logging.INFO,
    )
    logging_arguments.add_argument(
        "-vv",
        "--debug",
        help="Print lots of debugging statements. Sets logging level to DEBUG.",
        action="store_const",
        dest="loglevel",
        const=logging.DEBUG,
    )
    logging_arguments.add_argument(
        "--quiet",
        help="Be quiet! Sets logging level to WARNING.",
        action="store_const",
        dest="loglevel",
        const=logging.WARNING,
    )

    logging_arguments.add_argument(
        "--logging-config-file",
        type=str,
        help="If set, the name of the logging configuration file will be set "
        "to the given name.",
    )


def add_remote_storage_param(
    parser: Union[argparse.ArgumentParser, argparse._ActionsContainer],
    required: bool = False,
) -> None:
    parser.add_argument(
        "--remote-storage",
        help="Remote storage which should be used to store/load the model. "
        f"Supported storages are: {RemoteStorageType.list()}. "
        "You can also provide your own implementation of the `Persistor` interface.",
        required=required,
        type=parse_remote_storage_arg,
    )


def add_remote_root_only_param(
    parser: argparse.ArgumentParser, required: bool = False
) -> None:
    parser.add_argument(
        "--remote-root-only",
        action="store_true",
        help="If set, models will be stored only at the root directory "
        "of the remote storage.",
        required=required,
    )


def parse_remote_storage_arg(value: str) -> StorageType:
    try:
        return parse_remote_storage(value)
    except ValueError as e:
        raise argparse.ArgumentTypeError(str(e))


class SkipYamlValidation(Enum):
    DOMAIN = "domain"

    @classmethod
    def list(cls) -> List[str]:
        return [e.value for e in SkipYamlValidation]


def add_skip_validation_flag(
    parser: Union[argparse.ArgumentParser, argparse._ActionsContainer],
) -> None:
    parser.add_argument(
        "--skip-yaml-validation",
        default=[],
        choices=SkipYamlValidation.list(),
        action="append",
        help="Skip YAML validation for selected parts of the training data.",
    )


def add_sub_agents_param(
    parser: Union[argparse.ArgumentParser, argparse._ActionsContainer],
) -> None:
    parser.add_argument(
        "--sub-agents",
        type=str,
        default=DEFAULT_SUB_AGENTS,
        help="Directory that specifies sub-agents to use (default: %(default)s).",
    )



import logging
from typing import Any, Dict, Text

from rasa.shared.core.constants import USER_INTENT_OUT_OF_SCOPE
from rasa.shared.nlu.constants import (
    ENTITIES,
    ENTITY_ATTRIBUTE_ROLE,
    ENTITY_ATTRIBUTE_TYPE,
    ENTITY_ATTRIBUTE_VALUE,
    INTENT,
    TEXT,
)
from rasa.shared.nlu.training_data.formats.readerwriter import JsonTrainingDataReader
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.nlu.training_data.training_data import TrainingData

logger = logging.getLogger(__name__)


class WitReader(JsonTrainingDataReader):
    def read_from_json(self, js: Dict[Text, Any], **kwargs: Any) -> TrainingData:
        """Loads training data stored in the WIT.ai data format."""
        training_examples = []

        for s in js["utterances"]:
            entities = s.get(ENTITIES)
            if entities is None:
                continue
            text = s.get(TEXT)
            # Out-of-scope WIT utterances won't have the intent field set,
            # and that's the reason why we set it to `USER_INTENT_OUT_OF_SCOPE` by
            # default.
            intent = s.get("intent", USER_INTENT_OUT_OF_SCOPE)

            for e in entities:
                entity_name = e["entity"]
                if ":" not in entity_name:
                    continue
                (name, role) = entity_name.rsplit(":", 1)
                e[ENTITY_ATTRIBUTE_TYPE] = name
                e[ENTITY_ATTRIBUTE_ROLE] = role
                e[ENTITY_ATTRIBUTE_VALUE] = e.pop("body", None)

            data = {}
            if intent:
                data[INTENT] = intent
            if entities is not None:
                data[ENTITIES] = entities
            data[TEXT] = text
            training_examples.append(Message(data=data))
        return TrainingData(training_examples)



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict

from rasa.shared.utils.yaml import read_config_file

ChannelsType = Dict[str, Dict[str, Any]]


class CredentialsConfig:
    def __init__(self, channels: ChannelsType, config_file_path: Path) -> None:
        self.channels = channels
        self.config_file_path = config_file_path

    @classmethod
    def load_from_file(cls, file_path: Path) -> CredentialsConfig:
        credentials_dict = read_config_file(file_path)
        return cls(credentials_dict, file_path)



from typing import Any, Callable, Dict, Optional

from mcp import Tool
from pydantic import BaseModel

from rasa.agents.constants import (
    TOOL_ADDITIONAL_PROPERTIES_KEY,
    TOOL_DESCRIPTION_KEY,
    TOOL_EXECUTOR_KEY,
    TOOL_INPUT_SCHEMA_KEY,
    TOOL_NAME_KEY,
    TOOL_PARAMETERS_KEY,
    TOOL_PROPERTIES_KEY,
    TOOL_PROPERTY_TYPE_KEY,
    TOOL_REQUIRED_KEY,
    TOOL_STRICT_KEY,
    TOOL_TYPE_FUNCTION_KEY,
    TOOL_TYPE_KEY,
)


class AgentToolSchema(BaseModel):
    name: str
    parameters: dict[str, Any]
    strict: bool
    description: Optional[str] = None
    type: str = "function"

    @classmethod
    def from_mcp_tool(cls, tool: Tool) -> "AgentToolSchema":
        """Convert MCP Tool to AgentToolSchema."""
        parameters = tool.inputSchema.copy() if tool.inputSchema else {}

        if parameters:
            cls._validate_and_fix_parameters(parameters)

        return cls(
            name=tool.name,
            description=tool.description,
            parameters=parameters,
            strict=False,
            type=TOOL_TYPE_FUNCTION_KEY,
        )

    @classmethod
    def from_litellm_json_format(cls, tool: Dict[str, Any]) -> "AgentToolSchema":
        """Convert OpenAI dict format to AgentToolSchema."""
        references = (
            "Refer: \n"
            "- LiteLLM JSON Format - https://docs.litellm.ai/docs/completion/function_call#full-code---parallel-function-calling-with-gpt-35-turbo-1106\n"
            "- OpenAI Tool JSON Format - https://platform.openai.com/docs/guides/tools?tool-type=function-calling\n"
        )
        expected_structure = (
            "{\n"
            "    'type': 'function',\n"
            "    'function': {\n"
            "        'name': 'string',\n"
            "        'description': 'string',\n"
            "        'parameters': {\n"
            "            'type': 'object',\n"
            "            'properties': {'string': 'string'},\n"
            "            'required': ['string']\n"
            "        }\n"
            "    }\n"
            "}"
        )
        if (
            TOOL_NAME_KEY in tool
            and TOOL_DESCRIPTION_KEY in tool
            and TOOL_INPUT_SCHEMA_KEY in tool
        ):
            raise ValueError(
                "Anthropic Tool format is not supported yet. Please use the LiteLLM "
                "Tool format, which is based on OpenAI's format.\n"
                "The expected structure is:\n"
                f"{expected_structure}\n"
                f"{references}"
            )
        if not (TOOL_TYPE_FUNCTION_KEY in tool and TOOL_TYPE_KEY in tool):
            raise ValueError(
                "Invalid tool format. Expected a dictionary with 'type' and "
                "'function' keys. Expected LiteLLM Tool format that is based on OpenAI "
                "format which has the following structure: \n"
                f"{expected_structure}\n"
                f"{references}"
            )

        function_data = tool[TOOL_TYPE_FUNCTION_KEY]

        if not (
            TOOL_NAME_KEY in function_data and TOOL_DESCRIPTION_KEY in function_data
        ):
            raise ValueError(
                "Invalid tool format. 'function' must contain 'name' and "
                "'description' keys. Expected LiteLLM Tool format that is based on "
                "OpenAI format which has the following structure: \n"
                f"{expected_structure}\n"
                f"{references}"
            )
        parameters = function_data.get(TOOL_PARAMETERS_KEY, {})

        if parameters:
            cls._validate_and_fix_parameters(parameters)

        return cls(
            name=function_data[TOOL_NAME_KEY],
            description=function_data[TOOL_DESCRIPTION_KEY],
            parameters=parameters,
            strict=function_data.get(TOOL_STRICT_KEY, False),
            type=tool[TOOL_TYPE_KEY],
        )

    @staticmethod
    def _validate_and_fix_parameters(parameters: Dict[str, Any]) -> None:
        """Validate and fix parameters to ensure they meet OpenAI function calling
        requirements."""
        if not parameters:
            return

        # Ensure additionalProperties is set at the top level
        if TOOL_ADDITIONAL_PROPERTIES_KEY not in parameters:
            parameters[TOOL_ADDITIONAL_PROPERTIES_KEY] = False

        # Ensure required is set at the top level
        if TOOL_REQUIRED_KEY not in parameters:
            parameters[TOOL_REQUIRED_KEY] = []

        # Ensure all properties have types, required field and additionalProperties
        if TOOL_PROPERTIES_KEY in parameters:
            AgentToolSchema._ensure_property_types(parameters)

    @staticmethod
    def _ensure_property_types(parameters: Dict[str, Any]) -> None:
        """Ensure all properties in parameters have a type defined and
        additionalProperties is set.

        Properties that use structural keywords ($ref, anyOf, oneOf, allOf, enum)
        are left untouched, as they naturally omit the type field and adding one
        would break the schema.
        """
        properties = parameters[TOOL_PROPERTIES_KEY]

        if not properties:
            return

        # Structural keywords that indicate the type is determined elsewhere
        structural_keywords = {"$ref", "anyOf", "oneOf", "allOf", "enum"}

        for _, prop_schema in properties.items():
            if not isinstance(prop_schema, dict):
                continue

            # Skip adding type if the property uses structural keywords
            # These schemas naturally omit type and adding one would break validation
            if any(keyword in prop_schema for keyword in structural_keywords):
                continue

            # Ensure the property has a type
            if TOOL_PROPERTY_TYPE_KEY not in prop_schema:
                prop_schema[TOOL_PROPERTY_TYPE_KEY] = "string"

            # If it's an object type, ensure additionalProperties and required
            # fields are set
            if prop_schema[TOOL_PROPERTY_TYPE_KEY] == "object":
                if TOOL_ADDITIONAL_PROPERTIES_KEY not in prop_schema:
                    prop_schema[TOOL_ADDITIONAL_PROPERTIES_KEY] = False

                # Ensure required key exists for object properties
                if TOOL_REQUIRED_KEY not in prop_schema:
                    prop_schema[TOOL_REQUIRED_KEY] = []

    def to_litellm_json_format(self) -> Dict[str, Any]:
        """Convert AgentToolSchema to OpenAI format."""
        # Ensure the schema is valid before conversion
        return {
            TOOL_TYPE_KEY: TOOL_TYPE_FUNCTION_KEY,
            TOOL_TYPE_FUNCTION_KEY: self.model_dump(exclude={TOOL_TYPE_KEY}),
        }


class CustomToolSchema(BaseModel):
    """A class that represents the schema of a custom agent tool."""

    tool_name: str
    tool_definition: AgentToolSchema
    tool_executor: Callable

    @classmethod
    def from_dict(cls, config: Dict[str, Any]) -> "CustomToolSchema":
        """Convert a custom tool config to CustomToolSchema."""
        agent_tool_schema = AgentToolSchema.from_litellm_json_format(config)
        if TOOL_EXECUTOR_KEY not in config:
            raise ValueError("Custom tool executor is required.")

        return cls(
            tool_name=agent_tool_schema.name,
            tool_definition=agent_tool_schema,
            tool_executor=config[TOOL_EXECUTOR_KEY],
        )



"""MCP utilities."""


def mcp_server_exists(mcp_server: str) -> bool:
    """Check if an MCP server exists in the configured endpoints.

    Args:
        mcp_server: The name of the MCP server to check.

    Returns:
        True if the MCP server exists, False otherwise.
    """
    from rasa.core.config.configuration import Configuration

    endpoints = Configuration.get_instance().endpoints
    if (mcp_server_list := endpoints.mcp_servers) is None:
        return False

    mcp_server_names = [server.name for server in mcp_server_list]
    return mcp_server in mcp_server_names



import importlib.util
import sys

import structlog

structlogger = structlog.get_logger()


def check_tensorflow_installation() -> None:
    """Check if TensorFlow is installed without proper Rasa extras."""
    # Check if tensorflow is available in the environment
    tensorflow_available = importlib.util.find_spec("tensorflow") is not None

    if not tensorflow_available:
        return

    # Check if any TensorFlow-related extras were installed
    # We do this by checking for packages that are only installed with nlu/full extras
    tensorflow_extras_indicators = [
        "tensorflow_text",  # Only in nlu/full extras
        "tensorflow_hub",  # Only in nlu/full extras
        "tf_keras",  # Only in nlu/full extras
    ]

    extras_installed = any(
        importlib.util.find_spec(pkg) is not None
        for pkg in tensorflow_extras_indicators
    )

    if tensorflow_available and not extras_installed:
        structlogger.warning(
            "installation_utils.tensorflow_installation",
            warning=(
                "TensorFlow is installed but Rasa was not installed with TensorFlow "
                "support, i.e. additional packages required to use NLU components "
                "have not been installed. For the most reliable setup, delete your "
                "current virtual environment, create a new one, and install Rasa "
                "again. Please follow the instructions at "
                "https://rasa.com/docs/pro/installation/python"
            ),
        )


def check_tensorflow_integrity() -> None:
    """Check if TensorFlow installation is corrupted or incomplete."""
    # Only check if tensorflow is available
    if importlib.util.find_spec("tensorflow") is None:
        return

    try:
        # Try to import tensorflow - this will fail if installation is corrupted
        import tensorflow as tf

        # Try to access a basic TensorFlow function
        _ = tf.constant([1, 2, 3])
    except Exception:
        # Simplified error message for all TensorFlow corruption issues
        structlogger.error(
            "installation_utils.tensorflow_integrity",
            issue=(
                "TensorFlow is installed but appears to be corrupted or incomplete. "
                "For the most reliable setup, delete your current virtual "
                "environment, create a new one, and install Rasa again. "
                "Please follow the instructions at "
                "https://rasa.com/docs/pro/installation/python"
            ),
        )
        sys.exit(1)


def check_rasa_availability() -> None:
    """Check if Rasa is installed and importable."""
    if importlib.util.find_spec("rasa") is None:
        structlogger.error(
            "installation_utils.rasa_availability",
            issue=(
                "Rasa is not installed in this environment. "
                "Please follow the instructions at "
                "https://rasa.com/docs/pro/installation/python"
            ),
        )
        sys.exit(1)

    try:
        _ = importlib.import_module("rasa")
    except Exception as e:
        structlogger.error(
            "installation_utils.rasa_availability",
            issue=(
                f"Rasa is installed but cannot be imported: {e!s}."
                f"Please follow the instructions at "
                f"https://rasa.com/docs/pro/installation/python"
            ),
        )
        sys.exit(1)


def check_for_installation_issues() -> None:
    """Check for all potential installation issues.

    Returns:
        List of warning messages for detected issues.
    """
    # Check if Rasa is available first
    check_rasa_availability()

    # Check TensorFlow integrity first (more critical)
    check_tensorflow_integrity()

    # Check for orphaned TensorFlow
    check_tensorflow_installation()



"""The module is primarily centered around the `FlowRetrieval` class which handles the
initialization, configuration validation, vector store management, and flow retrieval
logic. It integrates components for managing embeddings, vector stores, and
flow-specific templates, facilitating semantic search functionalities.
Key Features:
- Configurable embedding strategies for dialogue components.
- Seamless interaction with model storage and resource management.
- Supports dynamic loading and persistence of vector stores.
- Enables population of vector stores based on specified dialogue flows and domain
information.
- Implements flow retrieval mechanisms including semantic search based on dialogue
context.
Usage:
Interaction with this class typically involves creating an instance with a
configuration dict, model storage, and a resource reference, then using methods
like `populate`, `persist`, or dynamic retrieval methods to manage or utilize
flows within a conversational context.
"""

import importlib
from typing import Any, Dict, List, Optional, Text

import structlog
from jinja2 import Template
from langchain.docstore.document import Document
from langchain.schema.embeddings import Embeddings
from langchain_community.vectorstores.faiss import FAISS
from langchain_community.vectorstores.utils import DistanceStrategy

import rasa.shared.utils.io
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.constants import (
    EMBEDDINGS_CONFIG_KEY,
    OPENAI_PROVIDER,
    PROVIDER_CONFIG_KEY,
)
from rasa.shared.core.domain import Domain
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.exceptions import ProviderClientAPIException
from rasa.shared.nlu.constants import FLOWS_FROM_SEMANTIC_SEARCH, TEXT
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.providers.embedding._langchain_embedding_client_adapter import (
    _LangchainEmbeddingClientAdapter,
)
from rasa.shared.utils.health_check.embeddings_health_check_mixin import (
    EmbeddingsHealthCheckMixin,
)
from rasa.shared.utils.llm import (
    DEFAULT_OPENAI_EMBEDDING_MODEL_NAME,
    USER,
    allowed_values_for_slot,
    embedder_factory,
    resolve_model_client_config,
    tracker_as_readable_transcript,
)

DEFAULT_FLOW_DOCUMENT_TEMPLATE = importlib.resources.read_text(
    "rasa.dialogue_understanding.generator", "flow_document_template.jinja2"
)

FLOW_RETRIEVAL_CONFIG_FILE_NAME = "flow_retrieval_config.json"

DEFAULT_EMBEDDINGS_CONFIG = {
    PROVIDER_CONFIG_KEY: OPENAI_PROVIDER,
    "model": DEFAULT_OPENAI_EMBEDDING_MODEL_NAME,
}

MAX_FLOWS_FROM_SEMANTIC_SEARCH_KEY = "num_flows"
SHOULD_EMBED_SLOTS_KEY = "should_embed_slots"
TURNS_TO_EMBED_KEY = "turns_to_embed"

DEFAULT_MAX_FLOWS_FROM_SEMANTIC_SEARCH = 20
DEFAULT_TURNS_TO_EMBED = 1
DEFAULT_SHOULD_EMBED_SLOTS = True


structlogger = structlog.get_logger()


class FlowRetrieval(EmbeddingsHealthCheckMixin):
    @classmethod
    def get_default_config(cls) -> Dict[str, Any]:
        """The default config for the flow retrieval."""
        return {
            EMBEDDINGS_CONFIG_KEY: DEFAULT_EMBEDDINGS_CONFIG,
            MAX_FLOWS_FROM_SEMANTIC_SEARCH_KEY: DEFAULT_MAX_FLOWS_FROM_SEMANTIC_SEARCH,
            TURNS_TO_EMBED_KEY: DEFAULT_TURNS_TO_EMBED,
            SHOULD_EMBED_SLOTS_KEY: DEFAULT_SHOULD_EMBED_SLOTS,
        }

    def __init__(
        self,
        config: Dict[str, Any],
        model_storage: ModelStorage,
        resource: Resource,
    ):
        config = {**self.get_default_config(), **config}
        self.config = self.validate_config(config)
        self.config[EMBEDDINGS_CONFIG_KEY] = resolve_model_client_config(
            self.config.get(EMBEDDINGS_CONFIG_KEY), FlowRetrieval.__name__
        )
        self.vector_store: Optional[FAISS] = None
        self.flow_document_template = DEFAULT_FLOW_DOCUMENT_TEMPLATE
        self._model_storage = model_storage
        self._resource = resource

    @classmethod
    def validate_config(cls, config: Dict[Text, Any]) -> Dict[Text, Any]:
        if config[MAX_FLOWS_FROM_SEMANTIC_SEARCH_KEY] < 0:
            config[MAX_FLOWS_FROM_SEMANTIC_SEARCH_KEY] = (
                DEFAULT_MAX_FLOWS_FROM_SEMANTIC_SEARCH
            )
            structlogger.error(
                f"flow_retrieval.validate_config.{MAX_FLOWS_FROM_SEMANTIC_SEARCH_KEY}.set_as_negative",
                event_info=(
                    f"The `{MAX_FLOWS_FROM_SEMANTIC_SEARCH_KEY}` is a "
                    f"negative value. Setting it to the default value "
                    f"({DEFAULT_MAX_FLOWS_FROM_SEMANTIC_SEARCH})."
                ),
                old_value=config[MAX_FLOWS_FROM_SEMANTIC_SEARCH_KEY],
                new_value=DEFAULT_MAX_FLOWS_FROM_SEMANTIC_SEARCH,
            )

        if config[TURNS_TO_EMBED_KEY] < 1:
            config[TURNS_TO_EMBED_KEY] = DEFAULT_TURNS_TO_EMBED
            structlogger.error(
                f"flow_retrieval.validate_config.{TURNS_TO_EMBED_KEY}.less_than_one",
                event_info=(
                    f"The `{TURNS_TO_EMBED_KEY}` is less than 1."
                    f"Setting it to the default value ({DEFAULT_TURNS_TO_EMBED})."
                    f"Only the latest user utterance will be used."
                ),
                old_value=config[TURNS_TO_EMBED_KEY],
                new_value=DEFAULT_TURNS_TO_EMBED,
            )

        return config

    @classmethod
    def load(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        **kwargs: Any,
    ) -> "FlowRetrieval":
        """Load flow retrieval with previously populated FAISS vector store."""

        # Perform health check on resolved embedding client config
        embeddings_config = resolve_model_client_config(
            config.get(EMBEDDINGS_CONFIG_KEY, {})
        )
        cls.perform_embeddings_health_check(
            embeddings_config,
            DEFAULT_EMBEDDINGS_CONFIG,
            "flow_retrieval.load",
            FlowRetrieval.__name__,
        )

        # initialize base flow retrieval
        flow_retrieval = FlowRetrieval(config, model_storage, resource)
        # load vector store
        vector_store = cls._load_vector_store(
            flow_retrieval.config, model_storage, resource
        )
        flow_retrieval.vector_store = vector_store

        return flow_retrieval

    @classmethod
    def _load_vector_store(
        cls, config: Dict[Text, Any], model_storage: ModelStorage, resource: Resource
    ) -> Optional[FAISS]:
        """Loads a FAISS vector store from a specified local path."""
        embeddings = cls._create_embedder(config)
        try:
            with model_storage.read_from(resource) as model_path:
                return FAISS.load_local(
                    folder_path=model_path,
                    embeddings=embeddings,
                    distance_strategy=DistanceStrategy.MAX_INNER_PRODUCT,
                    allow_dangerous_deserialization=True,
                )
        except Exception as e:
            structlogger.warning(
                "flow_retrieval.load_vector_store.failed",
                error=e,
                resource=resource.name,
            )
            return None

    @classmethod
    def _create_embedder(cls, config: Dict[Text, Any]) -> Embeddings:
        """Creates an embedder.

        Returns:
            The embedder.
        """
        # Copy the config so original config is not modified
        config = config.copy()
        # Resolve config and instantiate the embedding client
        config[EMBEDDINGS_CONFIG_KEY] = resolve_model_client_config(
            config.get(EMBEDDINGS_CONFIG_KEY), FlowRetrieval.__name__
        )
        client = embedder_factory(
            config.get(EMBEDDINGS_CONFIG_KEY), DEFAULT_EMBEDDINGS_CONFIG
        )
        # Wrap the embedding client in the adapter
        return _LangchainEmbeddingClientAdapter(client)

    def persist(self) -> None:
        self._persist_vector_store()
        self._persist_config()

    def _persist_vector_store(self) -> None:
        """Persists the FAISS vector store."""
        if self.vector_store is not None:
            with self._model_storage.write_to(self._resource) as model_path:
                self.vector_store.save_local(model_path)

    def _persist_config(self) -> None:
        with self._model_storage.write_to(self._resource) as path:
            rasa.shared.utils.io.dump_obj_as_json_to_file(
                path / FLOW_RETRIEVAL_CONFIG_FILE_NAME, self.config
            )

    def populate(self, flows: FlowsList, domain: Domain) -> None:
        """Populates the vector store with embeddings generated from
        documents based on the flow descriptions, and flow slots
        descriptions.

        Args:
            flows: List of flows to populate the vector store with.
            domain: The domain containing relevant slot information.
        """
        # Perform health check before populating the vector store with flows
        self.perform_embeddings_health_check(
            self.config.get(EMBEDDINGS_CONFIG_KEY),
            DEFAULT_EMBEDDINGS_CONFIG,
            "flow_retrieval.train",
            FlowRetrieval.__name__,
        )

        flows_to_embedd = flows.exclude_link_only_flows()

        if not flows_to_embedd:
            structlogger.debug(
                "flow_retrieval.populate_vector_store.no_flows_to_embed",
                event_info=(
                    "No flows to embed in the vector store, skipping population."
                ),
            )
            return

        embeddings = self._create_embedder(self.config)
        documents = self._generate_flow_documents(flows_to_embedd, domain)
        try:
            self.vector_store = FAISS.from_documents(
                documents=documents,
                embedding=embeddings,
                distance_strategy=DistanceStrategy.MAX_INNER_PRODUCT,
            )
        except Exception as e:
            error_type = e.__class__.__name__
            structlogger.error(
                "flow_retrieval.populate_vector_store.not_populated",
                event_info=(
                    "Failed to populate the FAISS store with the provided flows."
                ),
                error_type=error_type,
                error=e,
            )
            raise

    def _generate_flow_documents(
        self, flows: FlowsList, domain: Domain
    ) -> List[Document]:
        """Converts each flow in the provided list into an embeddable document. These
        documents include the following information for each flow: flow's name,
        flow's description, and associated slots and their descriptions and allowed
        values.

        Args:
            flows: List of flows.
            domain: The domain containing relevant slot information.

        Returns:
            List of documents, each representing the converted information of a single
            flow.
        """
        slots = {slot.name: slot for slot in domain.slots}
        flow_docs = []

        for flow in flows:
            flow_info: Dict[Text, Any] = {
                "flow": {
                    "name": flow.name,
                    "description": flow.description,
                    "slots": [],
                }
            }
            if self.config[SHOULD_EMBED_SLOTS_KEY]:
                flow_info["flow"]["slots"] = [
                    {
                        "name": q.collect,
                        "description": q.description,
                        "allowed_values": allowed_values_for_slot(slots[q.collect]),
                    }
                    for q in flow.get_collect_steps(deduplicate=True)
                ]

            flow_docs.append(
                Document(
                    page_content=Template(self.flow_document_template)
                    .render(flow_info)
                    .strip(),
                    metadata={"flow_id": flow.id},
                )
            )

        return flow_docs

    async def filter_flows(
        self, tracker: DialogueStateTracker, message: Message, flows: FlowsList
    ) -> FlowsList:
        """Filters the given flows.
        The filtered flows with the following rules:
        - Exclude flows that reachable only via link
        - Include flows set to be always included in the prompt
        - Include flows started during the conversation,
        - (Optionally) The rest of flows are filter so only the top 'k' with
          the highest similarity to the ongoing conversation are included.
          Set 'max_flows_from_semantic_search' to negative number to disable
          the option.

        Args:
            tracker: the tracker
            message: the user message
            flows: available flows

        Returns:
            List of flows that includes always-included flows, previously started flows
            and top `k` relevant flows for the current conversation.
        """
        # apply basic filtering
        flows = flows.exclude_link_only_flows()
        always_included_flows = flows.get_flows_always_included_in_prompt()
        previously_started_flows = tracker.get_previously_started_flows(flows)
        # apply semantic search filtering
        most_similar_flows = await self.find_most_similar_flows(tracker, message, flows)
        return FlowsList.from_multiple_flows_lists(
            always_included_flows,
            previously_started_flows,
            most_similar_flows,
        )

    async def find_most_similar_flows(
        self, tracker: DialogueStateTracker, message: Message, flows: FlowsList
    ) -> FlowsList:
        """Filters the given flows so only the top 'k' most similar
        flows to the current conversation are left.
        retrieved from the vector store.

        Args:
            tracker: the tracker
            message: the user message

        Returns:
            The most similar flows to the current conversation.
        """
        query = self._prepare_query(tracker, message)
        documents_with_scores = await self._query_vector_store(query)
        # filter out None i.e. more flows were embedded during training than are
        # available during prediction
        most_similar_flows_with_scores = [
            (flow, score)
            for doc, score in documents_with_scores
            if (flow := flows.flow_by_id(doc.metadata["flow_id"])) is not None
        ]
        # sort by score decreasing
        most_similar_flows_with_scores.sort(key=lambda x: x[1], reverse=True)
        # add the relevant flows to the message for evaluation purposes
        message.set(
            prop=FLOWS_FROM_SEMANTIC_SEARCH,
            info=[
                (flow.id, float(score))
                for flow, score in most_similar_flows_with_scores
            ],
            add_to_output=True,
        )
        return FlowsList([f for f, _ in most_similar_flows_with_scores])

    def _prepare_query(self, tracker: DialogueStateTracker, message: Message) -> Text:
        """Prepares the query for vector store. The query is composed
        of the conversation within the specified number of turns.

        Args:
            tracker: the tracker
            message: the user message

        Returns:
            The query for vector store.
        """
        if int(self.config[TURNS_TO_EMBED_KEY]) > 1:
            current_conversation = tracker_as_readable_transcript(
                tracker,
                human_prefix=USER,
                max_turns=int(self.config[TURNS_TO_EMBED_KEY]),
            )
            current_conversation += f"\n{USER}: {message.data[TEXT]}"
            return current_conversation

        return f"{message.data[TEXT]}"

    async def _query_vector_store(self, query: Text) -> List:
        """Compares the query with all flows using a vector store
        and returns the top k relevant flows for the current conversation.

        Args:
            query: query

        Returns:
            The top k documents with similarity scores.
        """
        if self.vector_store is None:
            return []
        try:
            documents_with_scores = (
                await self.vector_store.asimilarity_search_with_score(
                    query, k=int(self.config[MAX_FLOWS_FROM_SEMANTIC_SEARCH_KEY])
                )
            )
            structlogger.debug(
                "flow_retrieval.query_vector_store.fetched",
                event_info=(
                    f"Fetched the top {self.config[MAX_FLOWS_FROM_SEMANTIC_SEARCH_KEY]}"
                    f"similar flows from the vector store"
                ),
                query=query,
                top_k=self.config[MAX_FLOWS_FROM_SEMANTIC_SEARCH_KEY],
                results=[
                    {
                        "flow_id": document.metadata["flow_id"],
                        "score": score,
                        "content": document.page_content,
                    }
                    for document, score in documents_with_scores
                ],
            )
            return documents_with_scores
        except Exception as e:
            error_type = e.__class__.__name__
            structlogger.error(
                "flow_retrieval.query_vector_store.error",
                event_info="Cannot fetch flows from vector store",
                error_type=error_type,
                error=e,
                query=query,
            )
            raise ProviderClientAPIException(
                message="Cannot fetch flows from vector store", original_exception=e
            )



import os
from typing import Any, Dict, Optional

import structlog

from rasa.shared.constants import (
    OPENAI_API_BASE_ENV_VAR,
    OPENAI_API_TYPE_ENV_VAR,
    OPENAI_API_VERSION_ENV_VAR,
    OPENAI_PROVIDER,
)
from rasa.shared.providers._configs.openai_client_config import OpenAIClientConfig
from rasa.shared.providers.embedding._base_litellm_embedding_client import (
    _BaseLiteLLMEmbeddingClient,
)

structlogger = structlog.get_logger()


class OpenAIEmbeddingClient(_BaseLiteLLMEmbeddingClient):
    """A client for interfacing with OpenAI Embeddings.

    Parameters:
        model (str): The OpenAI model name.
        api_base (Optional[str]): Optional, the base URL for the API endpoints.
            If not provided, it will be set via environment variables.
        api_type (Optional[str]): Optional, the type of the API to use.
            If not provided, it will be set via environment variable.
        api_version (Optional[str]): Optional, the version of the API to use.
            If not provided, it will be set via environment variable.
        kwargs (Optional[Dict[str, Any]]): Optional configuration parameters specific
            to the embedding model deployment.

    Raises:
        ProviderClientValidationError: If validation of the client setup fails.
    """

    def __init__(
        self,
        model: str,
        api_base: Optional[str] = None,
        api_type: Optional[str] = None,
        api_version: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__()  # type: ignore
        self._model = model
        self._api_base = api_base or os.environ.get(OPENAI_API_BASE_ENV_VAR)
        self._api_type = api_type or os.environ.get(OPENAI_API_TYPE_ENV_VAR)
        self._api_version = api_version or os.environ.get(OPENAI_API_VERSION_ENV_VAR)
        self._extra_parameters = kwargs or {}
        self.validate_client_setup()

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "OpenAIEmbeddingClient":
        """
        Initializes the client from given configuration.

        Args:
            config (Dict[str, Any]): Configuration.

        Raises:
            ValueError: Raised in cases of invalid configuration:
                - If any of the required configuration keys are missing.
                - If `api_type` has a value different from `azure`.

        Returns:
            OpenAIEmbeddingClient: Initialized client.
        """
        try:
            openai_config = OpenAIClientConfig.from_dict(config)
        except ValueError as e:
            message = "Cannot instantiate a client from the passed configuration."
            structlogger.error(
                "openai_embedding_client.from_config.error",
                message=message,
                config=config,
                original_error=e,
            )
            raise

        return cls(
            model=openai_config.model,
            api_base=openai_config.api_base,
            api_type=openai_config.api_type,
            api_version=openai_config.api_version,
            **openai_config.extra_parameters,
        )

    @property
    def config(self) -> Dict:
        """
        Returns the configuration for the openai embedding client.

        Returns:
            Dictionary containing the configuration.
        """
        config = OpenAIClientConfig(
            model=self.model,
            api_base=self.api_base,
            api_type=self.api_type,
            api_version=self.api_version,
            extra_parameters=self._extra_parameters,
        )
        return config.to_dict()

    @property
    def model(self) -> str:
        """
        Returns the model name for the openai embedding client.

        Returns:
            String representing the model name.
        """
        return self._model

    @property
    def api_base(self) -> Optional[str]:
        """
        Returns the base API URL for the openai embedding client.

        Returns:
            String representing the base API URL.
        """
        return self._api_base

    @property
    def api_type(self) -> Optional[str]:
        """
        Returns the API type for the openai embedding client.

        Returns:
            String representing the API version.
        """
        return self._api_type

    @property
    def api_version(self) -> Optional[str]:
        """
        Returns the API version for the openai embedding client.

        Returns:
            String representing the API version.
        """
        return self._api_version

    @property
    def _litellm_extra_parameters(self) -> Dict[str, Any]:
        """
        Returns the model parameters for the openai embedding client.

        Returns:
            Dictionary containing the model parameters.
        """
        return self._extra_parameters

    @property
    def _embedding_fn_args(self) -> Dict[str, Any]:
        return {
            "model": self._litellm_model_name,
            "api_base": self.api_base,
            "api_type": self.api_type,
            "api_version": self.api_version,
            **self._litellm_extra_parameters,
        }

    @property
    def _litellm_model_name(self) -> str:
        """Returns the LiteLLM model name for the openai embedding client."""
        if self._model and f"{OPENAI_PROVIDER}/" not in self._model:
            return f"{OPENAI_PROVIDER}/{self._model}"
        return self._model



import logging
import os
from typing import List, Optional, Text, Union

import rasa.shared.data
import rasa.shared.utils.io
from rasa.shared.core.domain import Domain
from rasa.shared.core.training_data.story_reader.story_reader import StoryReader
from rasa.shared.core.training_data.story_reader.yaml_story_reader import (
    YAMLStoryReader,
)
from rasa.shared.core.training_data.structures import StoryStep
from rasa.shared.data import YAML_FILE_EXTENSIONS

logger = logging.getLogger(__name__)


def _get_reader(filename: Text, domain: Domain) -> StoryReader:
    if rasa.shared.data.is_likely_yaml_file(filename):
        return YAMLStoryReader(domain, filename)
    else:
        # This is a use case for uploading the story over REST API.
        # The source file has a random name.
        return _guess_reader(filename, domain)


def _guess_reader(filename: Text, domain: Domain) -> StoryReader:
    if YAMLStoryReader.is_stories_file(filename):
        return YAMLStoryReader(domain, filename)

    raise ValueError(
        f"Failed to find a reader class for the story file `{filename}`. "
        f"Supported formats are "
        f"{', '.join(YAML_FILE_EXTENSIONS)}."
    )


def load_data_from_resource(
    resource: Union[Text], domain: Domain, exclusion_percentage: Optional[int] = None
) -> List["StoryStep"]:
    """Loads core training data from the specified folder.

    Args:
        resource: Folder/File with core training data files.
        domain: Domain object.
        exclusion_percentage: Identifies the percentage of training data that
                              should be excluded from the training.

    Returns:
        Story steps from the training data.
    """
    if not os.path.exists(resource):
        raise ValueError(f"Resource '{resource}' does not exist.")

    return load_data_from_files(
        rasa.shared.utils.io.list_files(resource), domain, exclusion_percentage
    )


def load_data_from_files(
    story_files: List[Text], domain: Domain, exclusion_percentage: Optional[int] = None
) -> List["StoryStep"]:
    """Loads core training data from the specified files.

    Args:
        story_files: List of files with training data in it.
        domain: Domain object.
        exclusion_percentage: Identifies the percentage of training data that
                              should be excluded from the training.

    Returns:
        Story steps from the training data.
    """
    story_steps = []

    for story_file in story_files:
        reader = _get_reader(story_file, domain)

        steps = reader.read_from_file(story_file)
        story_steps.extend(steps)

    if exclusion_percentage and exclusion_percentage != 100:
        import random

        idx = int(round(exclusion_percentage / 100.0 * len(story_steps)))
        random.shuffle(story_steps)
        story_steps = story_steps[:-idx]

    return story_steps



import asyncio
import copy
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, cast

import structlog

from rasa.builder.copilot.constants import ROLE_COPILOT, ROLE_USER
from rasa.builder.copilot.copilot_response_handler import CopilotResponseHandler
from rasa.builder.copilot.models import (
    CopilotChatMessage,
    CopilotContext,
    GeneratedContent,
    ResponseCategory,
)
from rasa.builder.guardrails.clients import GuardrailsClient
from rasa.builder.guardrails.models import (
    GuardrailRequestKey,
    GuardrailResponse,
)
from rasa.builder.guardrails.utils import (
    create_guardrail_request,
)
from rasa.builder.shared.tracker_context import (
    AssistantConversationTurn,
    TrackerContext,
)

if TYPE_CHECKING:
    pass


structlogger = structlog.get_logger()


class GuardrailsPolicyChecker:
    def __init__(self, guardrail_client: GuardrailsClient):
        self.guardrail_client = guardrail_client

    async def check_assistant_chat_for_policy_violations(
        self,
        tracker_context: TrackerContext,
        hello_rasa_user_id: Optional[str],
        hello_rasa_project_id: Optional[str],
        **kwargs: Any,
    ) -> TrackerContext:
        """Return a sanitised TrackerContext with unsafe turns removed.

        Only user messages are moderated - assistant messages are assumed safe.
        LRU cache is used, so each unique user text is checked once.

        Args:
            tracker_context: The TrackerContext containing conversation turns.
            hello_rasa_user_id: The user ID for the conversation.
            hello_rasa_project_id: The project ID for the conversation.
            **kwargs: Additional parameters for the guardrail request.

        Returns:
            TrackerContext with unsafe turns removed.
        """
        # Collect (turn_index, user_text) for all turns with a user message
        items = self._format_user_messages_to_assistant(
            tracker_context.conversation_turns
        )

        flagged_turns = await self._check_user_messages_for_violations(
            items,
            hello_rasa_user_id=hello_rasa_user_id,
            hello_rasa_project_id=hello_rasa_project_id,
            log_prefix="assistant_guardrails",
            **kwargs,
        )

        if not flagged_turns:
            return tracker_context

        structlogger.info(
            "guardrails_policy_checker.assistant_guardrails.turns_flagged",
            count=len(flagged_turns),
            turn_indices=sorted(flagged_turns),
        )

        # Build a TrackerContext with safe turns
        safe_turns: List[AssistantConversationTurn] = [
            turn
            for idx, turn in enumerate(tracker_context.conversation_turns)
            if idx not in flagged_turns
        ]

        new_tracker_context = copy.deepcopy(tracker_context)
        new_tracker_context.conversation_turns = safe_turns
        return new_tracker_context

    async def check_copilot_chat_for_policy_violations(
        self,
        context: CopilotContext,
        hello_rasa_user_id: Optional[str],
        hello_rasa_project_id: Optional[str],
        **kwargs: Any,
    ) -> Optional[GeneratedContent]:
        """Check the copilot chat history for guardrail policy violations.

        Only user messages are moderated – assistant messages are assumed safe.
        LRU cache is used, so each unique user text is checked once.

        Args:
            context: The CopilotContext containing the copilot chat history.
            hello_rasa_user_id: The user ID for the conversation.
            hello_rasa_project_id: The project ID for the conversation.
            **kwargs: Additional parameters for the guardrail request.

        Returns:
            Returns a default violation response if the system flags any user message,
            otherwise return None.
        """
        # Collect (index, text) for user messages; skip ones already marked as
        # violations
        items = self._format_user_messages_to_copilot(context.copilot_chat_history)

        flagged_user_indices = await self._check_user_messages_for_violations(
            items,
            hello_rasa_user_id=hello_rasa_user_id,
            hello_rasa_project_id=hello_rasa_project_id,
            log_prefix="copilot_guardrails",
            **kwargs,
        )

        self._annotate_flagged_user_messages(
            context.copilot_chat_history, flagged_user_indices
        )

        if not flagged_user_indices:
            return None

        # Identify the latest user message index in the current request
        last_user_idx: Optional[int] = None
        for i in range(len(context.copilot_chat_history) - 1, -1, -1):
            if getattr(context.copilot_chat_history[i], "role", None) == ROLE_USER:
                last_user_idx = i
                break

        # Remove flagged user messages and their next copilot messages
        indices_to_remove: Set[int] = set()
        total = len(context.copilot_chat_history)
        for uidx in flagged_user_indices:
            indices_to_remove.add(uidx)
            next_idx = uidx + 1
            if (
                next_idx < total
                and getattr(context.copilot_chat_history[next_idx], "role", None)
                == ROLE_COPILOT
            ):
                indices_to_remove.add(next_idx)

        # Apply sanitization
        filtered_copilot_chat_history = [
            msg
            for i, msg in enumerate(context.copilot_chat_history)
            if i not in indices_to_remove
        ]
        if len(filtered_copilot_chat_history) != len(context.copilot_chat_history):
            structlogger.info(
                "guardrails_policy_checker"
                ".copilot_guardrails"
                ".copilot_chat_history_sanitized",
                removed_indices=sorted(indices_to_remove),
                removed_messages=(
                    len(context.copilot_chat_history)
                    - len(filtered_copilot_chat_history)
                ),
                kept_messages=len(filtered_copilot_chat_history),
            )
            context.copilot_chat_history = filtered_copilot_chat_history

        # Block only if the latest user message in this request was flagged
        if last_user_idx is not None and last_user_idx in flagged_user_indices:
            return CopilotResponseHandler.respond_to_guardrail_policy_violations()

        # Otherwise proceed (following messages are respected)
        return None

    async def _check_user_messages_for_violations(
        self,
        items: List[Tuple[int, str]],
        hello_rasa_user_id: Optional[str],
        hello_rasa_project_id: Optional[str],
        log_prefix: str,
        **kwargs: Any,
    ) -> Set[int]:
        """Run guardrail checks for provided (index, user_text) pairs.

        Args:
            items: List of tuples containing (index, user_text) to check.
            hello_rasa_user_id: The user ID for the conversation.
            hello_rasa_project_id: The project ID for the conversation.
            log_prefix: Prefix for logging messages.
            **kwargs: Additional parameters for the guardrail request.

        Returns:
            A set of indices that were flagged by the guardrails.
        """
        if not items:
            return set()

        # 1) Group indices by logical request key (hashable by value)
        indices_by_key: Dict[GuardrailRequestKey, List[int]] = {}
        for idx, text in items:
            key = GuardrailRequestKey(
                user_text=text,
                hello_rasa_user_id=hello_rasa_user_id or "",
                hello_rasa_project_id=hello_rasa_project_id or "",
                # The client-specific request parameters go in the metadata
                # so that they can be used to create the request
                metadata=kwargs,
            )
            if not key.user_text:
                continue
            indices_by_key.setdefault(key, []).append(idx)

        if not indices_by_key:
            return set()

        # 2) Create one task per logical key
        tasks_by_key: Dict[GuardrailRequestKey, asyncio.Task[GuardrailResponse]] = {}
        for key in indices_by_key:
            request = create_guardrail_request(
                client_type=type(self.guardrail_client),
                user_text=key.user_text,
                hello_rasa_user_id=key.hello_rasa_user_id,
                hello_rasa_project_id=key.hello_rasa_project_id,
                **kwargs,
            )
            tasks_by_key[key] = self.guardrail_client.schedule_check(request)

        # 3) Await unique tasks once
        keys = list(tasks_by_key.keys())
        tasks = [tasks_by_key[k] for k in keys]
        responses = await asyncio.gather(*tasks, return_exceptions=True)

        # 4) Map results back to all corresponding indices
        flagged: Set[int] = set()
        for key, response in zip(keys, responses):
            if isinstance(response, Exception):
                structlogger.warning(
                    f"{log_prefix}.request_failed", error=str(response)
                )
                continue
            # At this point, response is guaranteed to be GuardrailResponse
            # Use typing.cast to explicitly cast the type
            guardrail_response = cast(GuardrailResponse, response)
            if guardrail_response.flagged:
                flagged.update(indices_by_key.get(key, []))

        return flagged

    def _annotate_flagged_user_messages(
        self,
        copilot_chat_history: List[CopilotChatMessage],
        flagged_user_indices: Set[int],
    ) -> None:
        """Mark flagged user messages in-place on the original copilot chat history.

        Args:
            copilot_chat_history: The copilot chat history containing messages.
            flagged_user_indices: Set of indices of user messages that were flagged.
        """
        if not flagged_user_indices:
            return

        total = len(copilot_chat_history)
        for uidx in flagged_user_indices:
            if 0 <= uidx < total and copilot_chat_history[uidx].role == ROLE_USER:
                copilot_chat_history[
                    uidx
                ].response_category = ResponseCategory.GUARDRAILS_POLICY_VIOLATION

    def _format_user_messages_to_assistant(
        self, conversation_turns: List[AssistantConversationTurn]
    ) -> List[Tuple[int, str]]:
        """Collect (turn_index, user_text) tuples for all turns with user messages.

        Args:
            conversation_turns: The list of conversation turns.

        Returns:
            List of tuples containing (turn_index, user_text) for valid user messages.
        """
        items: List[Tuple[int, str]] = []
        for idx, turn in enumerate(conversation_turns):
            user_message = turn.user_message
            if not user_message:
                continue

            text = (user_message.text or "").strip()
            if not text:
                continue

            items.append((idx, text))

        return items

    def _format_user_messages_to_copilot(
        self, copilot_chat_history: List[CopilotChatMessage]
    ) -> List[Tuple[int, str]]:
        """Collect (index, user_text) tuples for all messages with user messages.

        Args:
            copilot_chat_history: The list of messages.
        """
        items: List[Tuple[int, str]] = []
        for idx, message in enumerate(copilot_chat_history):
            if (
                message.response_category
                == ResponseCategory.GUARDRAILS_POLICY_VIOLATION
            ):
                continue
            if message.role != ROLE_USER:
                continue
            formatted_message = message.build_openai_message()
            text = (formatted_message.get("content") or "").strip()
            if not text:
                continue
            items.append((idx, text))

        return items



import functools
import json
import time
from dataclasses import asdict, dataclass, field
from typing import Any, Awaitable, Callable, Dict, List, Optional, Text, Union

import structlog
from litellm.utils import ChatCompletionMessageToolCall
from pydantic import BaseModel

from rasa.shared.constants import KEY_TOOL_CALLS
from rasa.shared.exceptions import LLMToolResponseDecodeError

structlogger = structlog.get_logger()


@dataclass
class LLMUsage:
    prompt_tokens: int
    """Number of prompt tokens used to generate completion."""

    completion_tokens: int
    """Number of generated tokens."""

    total_tokens: int = field(init=False)
    """Total number of used tokens."""

    def __post_init__(self) -> None:
        self.total_tokens = self.prompt_tokens + self.completion_tokens

    @classmethod
    def from_dict(cls, data: Dict[Text, Any]) -> "LLMUsage":
        """Creates an LLMUsage object from a dictionary.
        If any keys are missing, they will default to zero
        or whatever default you prefer.
        """
        return cls(
            prompt_tokens=data.get("prompt_tokens"),
            completion_tokens=data.get("completion_tokens"),
        )

    def to_dict(self) -> dict:
        """Converts the LLMUsage dataclass instance into a dictionary."""
        return asdict(self)


class LLMToolCall(BaseModel):
    """A class representing a response from an LLM tool call."""

    id: str
    """The ID of the tool call."""

    tool_name: str
    """The name of the tool that was called."""

    tool_args: Dict[str, Any]
    """The arguments passed to the tool call."""

    type: str = "function"
    """The type of the tool call."""

    @classmethod
    def from_dict(cls, data: Dict[Text, Any]) -> "LLMToolCall":
        """Creates an LLMToolResponse from a dictionary."""
        return cls(**data)

    @classmethod
    def from_litellm(cls, data: ChatCompletionMessageToolCall) -> "LLMToolCall":
        """Creates an LLMToolResponse from a dictionary."""
        try:
            tool_args = json.loads(data.function.arguments)
        except json.JSONDecodeError as e:
            structlogger.error(
                "llm_response.litellm_tool_call.invalid_arguments",
                tool_name=data.function.name,
                tool_call=data.function.arguments,
            )
            raise LLMToolResponseDecodeError(
                original_exception=e,
                message=(
                    f"Invalid arguments for tool call - `{data.function.name}`: "
                    f"`{data.function.arguments}`"
                ),
            ) from e

        return cls(
            id=data.id,
            tool_name=data.function.name,
            tool_args=tool_args,
            type=data.type,
        )


@dataclass
class LLMResponse:
    id: str
    """A unique identifier for the completion."""

    choices: List[str]
    """The list of completion choices the model generated for the input prompt."""

    created: int
    """The Unix timestamp (in seconds) of when the completion was created."""

    model: Optional[str] = None
    """The model used for completion."""

    usage: Optional[LLMUsage] = None
    """An optional details about the token usage for the API call."""

    additional_info: Optional[Dict] = None
    """Optional dictionary for storing additional information related to the
    completion that may not be covered by other fields."""

    latency: Optional[float] = None
    """Optional field to store the latency of the LLM API call."""

    tool_calls: Optional[List[LLMToolCall]] = None
    """The list of tool calls the model generated for the input prompt."""

    @classmethod
    def from_dict(cls, data: Dict[Text, Any]) -> "LLMResponse":
        """Creates an LLMResponse from a dictionary."""
        usage_data = data.get("usage", {})
        usage_obj = LLMUsage.from_dict(usage_data) if usage_data else None

        tool_calls_data = data.get(KEY_TOOL_CALLS, [])
        tool_calls_obj = (
            [LLMToolCall.from_dict(tool) for tool in tool_calls_data]
            if tool_calls_data
            else None
        )

        return cls(
            id=data["id"],
            choices=data["choices"],
            created=data["created"],
            model=data.get("model"),
            usage=usage_obj,
            additional_info=data.get("additional_info"),
            latency=data.get("latency"),
            tool_calls=tool_calls_obj,
        )

    @classmethod
    def ensure_llm_response(cls, response: Union[str, "LLMResponse"]) -> "LLMResponse":
        if isinstance(response, LLMResponse):
            return response

        structlogger.warn("llm_response.deprecated_response_type", response=response)
        data = {"id": None, "choices": [response], "created": None}
        return LLMResponse.from_dict(data)

    def to_dict(self) -> dict:
        """Converts the LLMResponse dataclass instance into a dictionary."""
        result = asdict(self)
        if self.usage:
            result["usage"] = self.usage.to_dict()
        if self.tool_calls:
            result[KEY_TOOL_CALLS] = [tool.model_dump() for tool in self.tool_calls]
        return result


def measure_llm_latency(
    func: Callable[..., Awaitable[Optional[LLMResponse]]],
) -> Callable[..., Awaitable[Optional[LLMResponse]]]:
    @functools.wraps(func)
    async def wrapper(*args: Any, **kwargs: Any) -> Optional[LLMResponse]:
        start = time.perf_counter()
        result: Optional[LLMResponse] = await func(*args, **kwargs)
        if result:
            result.latency = time.perf_counter() - start
        return result

    return wrapper
